# Copyright 2024 Biomedical Data Science Lab, Universitat Politècnica de València (Spain)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Data Source Map main module.
"""

import numpy as np
import pandas as pd

from dataclasses import dataclass
from typing import Optional, List, Union, Dict

from dashi._constants import VALID_TEMPORAL_PERIODS, VALID_TYPES, VALID_STRING_TYPE, VALID_CATEGORICAL_TYPE, \
    VALID_INTEGER_TYPE, VALID_FLOAT_TYPE, \
    VALID_DATE_TYPE, TEMPORAL_PERIOD_WEEK, TEMPORAL_PERIOD_MONTH, TEMPORAL_PERIOD_YEAR, VALID_CONVERSION_STRING_TYPE, \
    MISSING_VALUE, VALID_TYPES_WITHOUT_DATE, VALID_DIM_REDUCTION_TYPES, PCA, MCA, FAMD
from dashi.unsupervised_characterization.utils import (_estimate_absolute_frequencies, _create_supports, _get_types,
                                                       BaseMultiVariateMap, _perform_dimensionality_reduction,
                                                       _scatter_plot, _compute_kde, _normalize_kde, _date_to_numeric)


@dataclass
class DataSourceMap:
    """
    A class that  contains the statistical distributions of data estimated at a
    specific time period. Both relative and absolute frequencies are included

    Attributes
    ----------
    probability_map: Union[List[List[float]], None]
        Numerical matrix representing the probability distribution temporal map (relative frequency).

    counts_map: Union[List[List[int]], None]
        Numerical matrix representing the counts temporal map (absolute frequency).

    sources: Union[List[str], None]
        List of sources (character) from which the data was obtained.

    support: Union[List[str], None]
        Numerical or character matrix representing the support (the value at each bin) of probability_map
        and counts_map.

    variable_name: Union[str, None]
        Name of the variable (character).

    variable_type: Union[str, None]
        Type of the variable (character).

    period: Union[str, None]
        Batching period among 'week', 'month' and 'year'.
    """

    probability_map: Union[List[List[float]], None] = None
    counts_map: Union[List[List[int]], None] = None
    sources: Union[List[str], None] = None
    support: Union[List[str], None] = None
    variable_name: Union[str, None] = None
    variable_type: Union[str, None] = None

    def check (self) -> Union[List[str], bool]:
        """
        Validates the consistency of the DataSourceMap attributes. This method checks for various
        potential issues, such as mismatched dimensions, invalid periods, or unsupported variable types.

        Returns
        -------
        Union[List[str], bool]:
            Returns a list of error messages if any validation fails, otherwise returns True indicating
            the object is valid.
        """
        errors = []

        # Check if the dimensions of probability_map and counts_map match
        if self.probability_map is not None and self.counts_map is not None:
            if (len(self.probability_map) != len(self.counts_map)
                    or any(len(probability_row) != len(count_row) for probability_row, count_row in
                           zip(self.probability_map, self.counts_map))):
                errors.append("the dimensions of probability_map and counts_map do not match")

        # Check if the length of dates matches the rows of probability_map
        if self.sources is not None and self.probability_map is not None:
            if len(self.sources) != len(self.probability_map):
                errors.append("the length of sources must match the rows of probability_map")

        # Check if the length of dates matches the rows of counts_map
        if self.sources is not None and self.counts_map is not None:
            if len(self.sources) != len(self.counts_map):
                errors.append("the length of sources must match the rows of counts_map")

        # Check if the length of support matches the columns of probability_map
        if self.support is not None and self.probability_map is not None:
            if len(self.support) != len(self.probability_map):
                errors.append("the length of support must match the columns of probability_map")

        # Check if the length of support matches the columns of counts_map
        if self.support is not None and self.counts_map is not None:
            if len(self.support) != len(self.counts_map):
                errors.append("the length of support must match the columns of counts_map")

        # Check if period is one of the valid periods
        if self.period is not None and self.period not in VALID_TEMPORAL_PERIODS:
            errors.append(f"period must be one of the following: {', '.join(VALID_TEMPORAL_PERIODS)}")

        # Check if variableType is one of the valid types
        if self.variable_type is not None and self.variable_type not in VALID_TYPES:
            errors.append(f"variable_type must be one of the following: {', '.join(VALID_TYPES)}")

        return errors if errors else True

@dataclass
class MultiVariateDataSourceMap(BaseMultiVariateMap, DataSourceMap):
    """
    A subclass of DataSourceMap representing a multi-variate multi-source data map.
    In addition to the attributes inherited from the DataSourceMap class, this
    class includes additional properties specific to multivariate multi-source data.

    Attributes
    ----------
    multivariate_probability_map: Optional[np.ndarray]
        List of matrices representing the multi-variate probability distribution
        temporal map (relative frequency) for each timestamp.

    multivariate_counts_map: Optional[np.ndarray]
        List of matrices representing the multi-variate counts temporal map (absolute)
        for each timestamp.

    multivariate_support: Optional[np.ndarray]
        List of matrices representing the support (the value at each bin) of the dimensions
        of multivariate_probability_map and multivariate_counts_map.
    """
    def check(self):
        errors = super().check()
        multi_errors = self.check_multivariate()
        if multi_errors is not True:
            if errors is True:
                errors = []
            errors.extend(multi_errors)
        return errors if errors else True


def estimate_univariate_data_source_map(
    data: pd.DataFrame,
    source_column: str,
    supports: Union[Dict, None] = None,
    numeric_smoothing: bool = True,
    numeric_variables_bins: Optional[int] = 100,
    verbose: bool = False
) -> Union[DataSourceMap, Dict[str, DataSourceMap]]:
    """
    Estimates a DataSourceMap object from a DataFrame containing individuals in rows and the variables
    in columns, being one of these columns the analysis source.

    Parameters
    ----------
    data: pd.DataFrame
        DataFrame containing the data to be analyzed. Each row represents an individual and each column a variable.

    source_column: str
        Name of the column in the DataFrame that contains the source of the data. This column will be used to group
        the data and estimate the distributions.

    supports: Union[Dict, None], optional
        A dictionary with structure {variable_name: variable_type_name} containing the support
        of the data distributions for each variable. If not provided, it is automatically
        estimated from the data.

    numeric_smoothing: bool, optional
        Logical value indicating whether a Kernel Density Estimation smoothing
        (Gaussian kernel, default bandwidth) is to be applied on numerical variables
        or traditional histogram instead.

    numeric_variables_bins: int
        The number of bins at which to define the frequency/density histogram for numerical
        variables when their support is not provided. 100 as default.

    verbose: bool, optional
        If True, prints additional information about the estimation process. Default is False.

    Returns
    -------
    DataSourceMap
        The DataSourceMap object or a dictionary of DataSourceMap objects depending on the number of
        analysis variables.
"""

    if data is None:
        raise ValueError('An input data frame is required')

    if source_column is None:
        raise ValueError('A source column name is required')

    if source_column not in data.columns:
        raise ValueError(f'Source column "{source_column}" not found in the data frame')

    if not all(data[column].dtype.name in VALID_TYPES for column in data.columns):
        print(data.dtypes)
        raise ValueError(f'The classes of input columns must be one of the following: {", ".join(VALID_TYPES)}')

    if supports is not None and not all(support in VALID_TYPES_WITHOUT_DATE for support in supports):
        raise ValueError(
            f'All the elements provided in the supports parameter must be of type {", ".join(VALID_TYPES_WITHOUT_DATE)}')


    sources = data[source_column]
    data_without_sources = data.drop(columns=[source_column])
    number_of_columns = len(data_without_sources.columns)

    if number_of_columns == 0:
        raise ValueError("Data must contain at least one variable column apart from source.")

    if verbose:
        print(f'Total number of columns to analyze: {number_of_columns}')
        print(f'Number of sources: {len(sources.unique())}')

    # Get VARIABLE types, others will not be allowed
    data_types, columns_by_type = _get_types(
        data=data_without_sources,
        verbose=verbose
    )

    data_without_sources = _date_to_numeric(
        data=data_without_sources,
        columns_by_type=columns_by_type,
        verbose=verbose
    )

    # Implement the logic to fill supports based on the variable type
    # Create supports
    data_without_sources, supports = _create_supports(
        data=data_without_sources,
        supports=supports,
        columns_types=columns_by_type,
        number_of_columns=number_of_columns,
        numeric_variables_bins=numeric_variables_bins,
        verbose=verbose
    )

    posterior_data_types = data_without_sources.dtypes
    results: dict = {}

    for column in data_without_sources.columns:
        if verbose:
            print(f'Estimation of DataSourceMap for variable: {column}')

        grouped = data_without_sources.groupby(sources, observed=False)[column]

        counts_map:list = []

        for source_value, group in grouped:
            map_data = _estimate_absolute_frequencies(
                group,
                varclass=posterior_data_types[column],
                support=supports[column],
                numeric_smoothing=numeric_smoothing
            )
            counts_map.append(map_data)
        counts_map = np.array(counts_map)

        probability_map = np.array([
            arr / arr.sum() if arr.sum() > 0 else np.zeros_like(arr)
            for arr in counts_map
        ])

        if posterior_data_types[column] == VALID_DATE_TYPE:
            support = pd.DataFrame(pd.to_datetime(supports[column]))
        elif posterior_data_types[column] in [VALID_STRING_TYPE, VALID_CATEGORICAL_TYPE]:
            support = pd.DataFrame(supports[column], columns=[column])
        else:
            support = pd.DataFrame(supports[column])

        data_source_map = DataSourceMap(
            probability_map=probability_map,
            counts_map=counts_map,
            sources=np.unique(sources),
            support=support,
            variable_name=column,
            variable_type=posterior_data_types[column]
        )
        results[column] = data_source_map

    if number_of_columns > 1:
        if verbose:
            print('Returning results as a dictionary of DataSourceMap objects for each column')
        return results

    else:
        if verbose:
            print('Returning a single DataSourceMap object for the single column')
        return results[data_without_sources.columns[0]]


def estimate_multivariate_data_source_map(
        data: pd.DataFrame,
        source_column_name: str,
        kde_resolution: int = 10,
        dimensions: int = 2,
        dim_reduction: str = 'PCA',
        scale: bool = True,
        scatter_plot: bool = False,
        verbose: bool = False
) -> MultiVariateDataSourceMap:
    """
    Estimates a MultiVariateDataSourceMap object from a DataFrame containing multiple variables (in columns) with a source
    column indicating the source of the data. The function performs dimensionality reduction on the data (e.g. PCA, MCA, FAMD)
    to handle high-dimensional data and estimates the probability distributions for each source.

    Parameters
    ----------
    data: pd.DataFrame
        A DataFrame where each row represents an individual or data point, and each column represents a
        variable. One column should represent the analysis sources.

    source_column_name: str
        A string indicating the name of the column in the DataFrame that contains the source of the data.

    kde_resolution: int
        The resolution of the grid used for Kernel Density Estimation (KDE). This determines the granularity
        of the KDE grid and how fine or coarse the estimated density maps will be. Default is 10.

    dimensions: int
        The number of dimensions to keep after applying dimensionality reduction (e.g., PCA).
        Default is 2, meaning the data will be projected into a 2D space. The maximum number of dimensions
        available are 3.

    dim_reduction: str
        A dimensionality reduction technique to be used on the data. Default is `PCA` (Principal Component Analysis)
        for numerical data. Other options can include 'MCA' (Multiple Correspondence Analysis) for categorical data or
        'FAMD' (Factor Analysis of Mixed Data) for mixed data. Note: in case of using 'FAMD', numerical variables must be
        in float type. Otherwise they will be treated as categorical.

    scale: bool
        Applicable just when using PCA dimensionality reduction. If true scales the input data using z-score
        normalization. Defaults to `True`.

    scatter_plot: bool
        Whether to generate a scatter plot of the first two principal components of the dimensionality reduction

    verbose: bool
        Whether to display additional information during the process. Defaults to `False`.

    Returns
    -------
    MultiVariateDataSourceMap
        A MultiVariateDataSourceMap object containing the estimated probability distributions for each source,
        along with the multivariate probability maps, counts maps, and supports for each dimension.
    """
    if data is None:
        raise ValueError('An input data frame is required')

    if source_column_name is None:
        raise ValueError('A source column name is required')

    if source_column_name not in data.columns:
        raise ValueError(f'Source column "{source_column_name}" not found in the data frame')

    if not all(data[column].dtype.name in VALID_TYPES for column in data.columns):
        print(data.dtypes)
        raise ValueError(f'The classes of input columns must be one of the following: {", ".join(VALID_TYPES)}')

    if dim_reduction not in VALID_DIM_REDUCTION_TYPES:
        raise ValueError(f'Dimensionality reduction method must be one of the following: {", ".join(VALID_DIM_REDUCTION_TYPES)}')

    if dimensions not in [2, 3]:
        raise ValueError('Dimensions must be either 2 or 3 for the multivariate data source map')

    sources = data[source_column_name]
    data_without_sources = data.drop(columns=[source_column_name])
    number_of_columns = len(data_without_sources.columns)

    if verbose:
        print(f'Total number of columns to analyze: {number_of_columns}')
        print(f'Number of sources: {len(sources.unique())}')

    data_types, columns_by_type = _get_types(
        data=data_without_sources,
        verbose=verbose
    )

    data_without_sources = _date_to_numeric(
        data=data_without_sources,
        columns_by_type=columns_by_type,
        verbose=verbose
    )

    if verbose:
        print(f'Applying dimensionality reduction with {dim_reduction}')

    reduced_data = _perform_dimensionality_reduction(
        data_without_sources,
        dim_reduction=dim_reduction,
        n_components=dimensions,
        verbose=verbose,
        scale=scale
    )

    reduced_data[[source_column_name]] = pd.DataFrame({
        source_column_name: sources
    })

    if scatter_plot:
        _scatter_plot(reduced_data=reduced_data, dim_reduction=dim_reduction, verbose=verbose,
                      color_column=source_column_name)

    value_counts = sources.value_counts(sort=False)

    sources_info = {
        'sources': np.unique(sources),
        'value_counts': value_counts[value_counts > dimensions],
        'source_column_name': source_column_name
    }

    dsm = _generate_multivariate_dsm(
        reduced_data=reduced_data,
        sources_info=sources_info,
        verbose=verbose,
        dimensions=dimensions,
        kde_resolution=kde_resolution
    )

    return dsm

def estimate_conditional_data_source_map(
        data: pd.DataFrame,
        source_column_name: str,
        label_column_name: str,
        kde_resolution: int = 10,
        dimensions: int = 2,
        dim_reduction: str = 'PCA',
        scale: bool = True,
        scatter_plot: bool = False,
        verbose: bool = False
) -> Dict[str, MultiVariateDataSourceMap]:
    """
    Estimates a MultiVariateDataSourceMap object for the data corresponding to each label of a DataFrame
    containing multiple variables (in columns) over different sources, using dimensionality reduction techniques (e.g., PCA)
    to handle high dimensional data.

    Parameters
    ----------
    data: pd.DataFrame
        A DataFrame where each row represents an individual or data point, and each column represents a
        variable. One column should represent the analysis sources and another the labels.

    source_column_name: str
        A string indicating the name of the column in the DataFrame that contains the source of the data.

    label_column_name: str
        The name of the column that contains the labels or class/category for each observation
        (used for conditional analysis).

    kde_resolution: int
        The resolution of the grid used for Kernel Density Estimation (KDE). This determines the granularity
        of the KDE grid and how fine or coarse the estimated density maps will be. Default is 10.

    dimensions: int
        The number of dimensions to keep after applying dimensionality reduction (e.g., PCA).
        Default is 2, meaning the data will be projected into a 2D space. The maximum number of dimensions
        available are 3. For single variable datasets, dimensions can be set to 1

    dim_reduction: str
        A dimensionality reduction technique to be used on the data. Default is 'PCA' (Principal Component Analysis)
        for numerical data. Other options can include 'MCA' (Multiple Correspondence Analysis) for categorical data or
        'FAMD' (Factor Analysis of Mixed Data) for mixed data. Note: in case of using 'FAMD', numerical variables must be
        in float type. Otherwise they will be treated as categorical.

    scale: str
        Applicable just when using PCA dimensionality reduction. If true scales the input data using z-score
        normalization. Defaults to `True`

    scatter_plot: bool
        Whether to generate a scatter plot of the first two principal components of the dimensionality reduction.

    verbose: bool
        Whether to display additional information during the process. Defaults to `False`.

    Returns
    -------
    Dict[str, MultiVariateDataSourceMap]
        A dictionary where the keys are the labels in the dataset, and the values are
        `MultiVariateDataSourceMap` objects representing the multi-source maps generated for each label.
    """
    if data is None:
        raise ValueError('An input data frame is required')

    if source_column_name is None:
        raise ValueError('The name of the column including sources is required.')

    if source_column_name not in data.columns:
        raise ValueError(f'Source column "{source_column_name}" not found in the data frame')

    if not all(data[column].dtype.name in VALID_TYPES for column in data.columns):
        print(data.dtypes)
        raise ValueError(f'The types of input columns must be one of the following: {", ".join(VALID_TYPES)}')

    if dim_reduction not in VALID_DIM_REDUCTION_TYPES:
        raise ValueError(
            f'Dimensionality reduction method must be one of the following: {", ".join(VALID_DIM_REDUCTION_TYPES)}')

    if dimensions not in [1, 2, 3]:
        raise ValueError(
            f'The number of supported dimensions are 1, 2 or 3')

    labels_column = data[label_column_name]
    sources = data[source_column_name]
    data_without_sources = data.drop(columns=[source_column_name, label_column_name])
    number_of_columns = len(data_without_sources.columns)

    if verbose:
        print(f'Total number of columns to analyze: {number_of_columns}')
        print(f'Number of sources: {len(sources.unique())}')

    data_types, columns_by_type = _get_types(
        data=data_without_sources,
        verbose=verbose
    )

    data_without_sources = _date_to_numeric(
        data=data_without_sources,
        columns_by_type=columns_by_type,
        verbose=verbose
    )

    if verbose:
        print(f'Applying dimensionality reduction with {dim_reduction}')

    reduced_data = _perform_dimensionality_reduction(
        data_without_sources,
        dim_reduction=dim_reduction,
        n_components=dimensions,
        verbose=verbose,
        scale=scale
    )

    reduced_data[[label_column_name, source_column_name]] = pd.DataFrame({
        label_column_name: labels_column,
        source_column_name: sources
    })

    if scatter_plot:
        _scatter_plot(reduced_data=reduced_data, dim_reduction=dim_reduction, verbose=verbose,
                      color_column=label_column_name)

    reduced_data_by_label = {
        label: group.drop(columns=[label_column_name]).reset_index(drop=True)
        for label, group in reduced_data.groupby(label_column_name, observed=True)
    }

    concept_maps_dict: dict = {}
    for label, concept_data in reduced_data_by_label.items():
        if verbose:
            print(f'Label: {label}')

        value_counts = concept_data[source_column_name].value_counts(sort=False)
        sources_info = {
            'sources': np.unique(sources),
            'value_counts': value_counts[value_counts > dimensions],
            'source_column_name': source_column_name
        }
        dsm = _generate_multivariate_dsm(
            reduced_data=concept_data,
            sources_info=sources_info,
            verbose=verbose,
            dimensions=dimensions,
            kde_resolution=kde_resolution
        )

        concept_maps_dict[label] = dsm

    return concept_maps_dict


def _generate_multivariate_dsm(
        reduced_data,
        sources_info,
        verbose,
        dimensions,
        kde_resolution
):
    xmin = reduced_data.drop(columns=[sources_info['source_column_name']]).min(axis=0)
    xmax = reduced_data.drop(columns=[sources_info['source_column_name']]).max(axis=0)

    if verbose:
        print('Estimating the data source maps')

    kde_list: list = []
    for source in sources_info['sources']:
        if source in sources_info['value_counts'].index and sources_info['value_counts'][source] > dimensions:
            kde = _compute_kde(
                reduced_data[reduced_data[sources_info['source_column_name']] == source].drop(
                    columns=[sources_info['source_column_name']]
                ),
                xmin[:dimensions], xmax[:dimensions], kde_resolution
            )
            kde_list.append(kde)
        else:
            if verbose:
                print(f'Not enough data for calculating source: {source} probability map')
            kde_shape = (kde_resolution,) * dimensions
            kde = np.full(kde_shape, np.nan)
            kde_list.append(kde)

    probability_map = np.row_stack([_normalize_kde(kde).flatten() for kde in kde_list])
    multivariate_probability_map = [_normalize_kde(kde) for kde in kde_list]
    multivariate_support = [np.linspace(start, stop, kde_resolution) for start, stop in zip(xmin[:dimensions], xmax[:dimensions])]
    non_nan_mask = ~np.isnan(probability_map).any(axis=1)
    non_nan_probability_map = probability_map[non_nan_mask]
    non_nan_counts_map = np.round(non_nan_probability_map * sources_info['value_counts'].values[:, np.newaxis])
    counts_map = np.full(probability_map.shape, np.nan)
    counts_map[non_nan_mask] = non_nan_counts_map
    multivariate_counts_map: list = []
    index = 0
    for prob_map in multivariate_probability_map:
        if np.isnan(prob_map).any():
            multivariate_counts_map.append(prob_map)
        else:
            multivariate_counts_map.append(np.round(prob_map * sources_info['value_counts'].iloc[index]))
            index += 1

    dsm = MultiVariateDataSourceMap(
        probability_map=probability_map,
        multivariate_probability_map=multivariate_probability_map,
        counts_map=counts_map,
        multivariate_counts_map=multivariate_counts_map,
        sources=sources_info['sources'],
        support=pd.DataFrame(range(0, kde_resolution ** dimensions)),
        multivariate_support=multivariate_support,
        variable_name='Dim.reduced.{}D'.format(dimensions),
        variable_type='float64',
    )

    return dsm






# Copyright 2024 Biomedical Data Science Lab, Universitat Politècnica de València (Spain)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Main function for estimating models over multiple temporal or multi-source batches.
"""

# MODULES IMPORT
import warnings
from typing import List, Dict, Optional, Tuple

import sklearn.metrics as skmet
from numpy import ndarray, sqrt
from pandas import DataFrame, concat, to_datetime
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.model_selection import train_test_split, KFold
from sklearn.preprocessing import LabelEncoder, RobustScaler, OneHotEncoder
from tqdm.auto import tqdm


# FUNCTION DEFINITION
def estimate_multibatch_models(*, data: DataFrame, inputs_numerical_column_names: Optional[List[str]] = None,
                               inputs_categorical_column_names: Optional[List[str]] = None,
                               output_regression_column_name: Optional[str] = None,
                               output_classification_column_name: Optional[str] = None,
                               date_column_name: Optional[str] = None,
                               period: Optional[str] = None, source_column_name: Optional[str] = None,
                               learning_strategy: Optional[str] = 'from_scratch') -> Dict[Tuple, Dict[str, float]]:
    """
    Estimates models across multiple batches, based on either time (temporal) or source.
    Requires specifying one target variable (regression or classification) and at least one
    numerical or categorical input feature within the input DataFrame.
    At the same time, it is necessary to provide either a date variable (indicating the period with the corresponding
    argument) or a source variable. The date variable must be a valid date, and the source variable categories need to
    be specified as strings.
    Additionally, it is recommended that the amount of data in each batching group be  sufficient for statistical
    representativeness.

    Parameters
    ----------
    data : DataFrame
        The input data containing numerical and/or categorical features, as well as the target variable
        (either a classification or regression target).

    inputs_numerical_column_names : Optional[List[str]], default=None
        List of column names representing numerical input features. If there are no numerical input features,
        set this to None.

    inputs_categorical_column_names : Optional[List[str]], default=None
        List of column names representing categorical input features. If there are no categorical input features,
        set this to None.

    output_regression_column_name : Optional[str], default=None
        Column name for the regression target variable. If there is no regression target, set this to None.

    output_classification_column_name : Optional[str], default=None
        Column name for the classification target variable. If there is no classification target, set this to None.

    date_column_name : Optional[str], default=None
        Column name containing date or time information for temporal batching. If performing source-based
        analysis instead of temporal batching, set this to None.

    period : Optional[str], default=None
        Period for batching the data ('month' or 'year') when using temporal batching. If not using temporal
        batching, set this to None.

    source_column_name : Optional[str], default=None
        Column name representing the source of the data (for multi-source batching). If performing temporal
        batching, set this to None.

    learning_strategy : Optional[str], default='from_scratch'
        Defines the learning strategy: either 'from_scratch' or 'cumulative'. Note that the 'cumulative' strategy
        can only be applied to temporal analyses, not multi-source analyses.

    Returns
    -------
    Dict[Tuple, Dict[str, float]]
        A dictionary where each key is a tuple `(train_batch_ids, test_batch_id, 'test')`
        representing the training/testing combination. Each corresponding value is another
        dictionary containing the calculated performance metrics for that specific test.

        The inner dictionary contains metric names (str) and their values (float):
        Regression metrics, if applicable:
            - 'MEAN_ABSOLUTE_ERROR'
            - 'MEAN_SQUARED_ERROR'
            - 'ROOT_MEAN_SQUARED_ERROR'
            - 'R_SQUARED'
        Classification metrics, if applicable:
            - 'AUC_{class_identifier}'
            - 'AUC_MACRO'
            - 'LOGLOSS'
            - 'RECALL_{class_identifier}'
            - 'PRECISION_{class_identifier}'
            - 'F1-SCORE_{class_identifier}'
            - 'ACCURACY'
            - 'RECALL_MACRO'
            - 'RECALL_MICRO'
            - 'RECALL_WEIGHTED'
            - 'PRECISION_MACRO'
            - 'PRECISION_MICRO'
            - 'PRECISION_WEIGHTED'
            - 'F1-SCORE_MACRO'
            - 'F1-SCORE_MICRO'
            - 'F1-SCORE_WEIGHTED'
    """

    # Input checking
    _check_inputs(data=data, inputs_numerical_column_names=inputs_numerical_column_names,
                  inputs_categorical_column_names=inputs_categorical_column_names,
                  output_regression_column_name=output_regression_column_name,
                  output_classification_column_name=output_classification_column_name,
                  date_column_name=date_column_name, period=period, source_column_name=source_column_name,
                  learning_strategy=learning_strategy)

    # Memory allocation for metrics
    metrics = {}

    # Modeling settings
    number_trees = 450
    maximum_depth = 9
    random_seed = 42

    # Make a copy to avoid modifying the original DataFrame
    data = data.copy()

    # Label encoding (for classification tasks)
    if output_classification_column_name is not None:
        # label encoder initialization
        label_encoder = LabelEncoder()
        # label encoding
        encoded = label_encoder.fit_transform(data[output_classification_column_name])

        # Replace column with integer dtype
        data[output_classification_column_name] = encoded.astype("int64")
        # index to class map derivation
        index2class_map = dict(enumerate(label_encoder.classes_))

    # Batching logic: temporal or multi-source
    # multi-source analysis
    if date_column_name is None and source_column_name is not None:
        batching_column_name = source_column_name
    # temporal analysis
    elif date_column_name is not None and source_column_name is None:
        # convert column to datetime format if not already, handling both strings and datetimes
        data[date_column_name] = to_datetime(data[date_column_name])
        # sorting by date
        data = data.sort_values(by=date_column_name)
        # batching period adjusting
        if period == 'month':
            data[date_column_name] = data[date_column_name].dt.strftime("%B %Y")
        elif period == 'year':
            data[date_column_name] = data[date_column_name].dt.strftime("%Y")
        else:
            raise ValueError("Current supported batching periods are 'month' and 'year'.")
        # batching column assignation
        batching_column_name = date_column_name
    else:
        raise ValueError("Invalid configuration: Please provide exactly one of 'date_column_name' for "
                         "temporal batching or 'source_column_name' for source-based batching. "
                         "You have either provided both or neither.")

    # Generate split indexes based on batching
    split_indexes = _generate_split_indexes(data=data, batching_column_name=batching_column_name)

    # Extract batch identifiers
    batch_identifiers = tuple(split_indexes.keys())

    # Memory allocation
    trained_scalers = {}
    trained_encoders = {}
    trained_models = {}

    # Generate combinations for training and testing batches
    # memory allocation
    combinations = []
    # filling
    if learning_strategy == 'from_scratch':
        for batch_idf_train in batch_identifiers:
            combinations.append(((batch_idf_train,), batch_idf_train, 'train',))
            for batch_idf_test in batch_identifiers:
                combinations.append(((batch_idf_train,), batch_idf_test, 'test'))
    elif learning_strategy == 'cumulative':
        for idx, batch_idf_train in enumerate(batch_identifiers):
            batch_cumulative_identifiers = tuple([batch_identifiers[i] for i in range(0, idx + 1)])
            combinations.append((batch_cumulative_identifiers, batch_idf_train, 'train',))
            for batch_idf_test in batch_identifiers:
                combinations.append((batch_cumulative_identifiers, batch_idf_test, 'test'))
    else:
        raise ValueError('Unrecognized learning strategy.')

    # Preprocessing, training and evaluation
    for combination in tqdm(combinations, total=len(combinations), colour='#32CD32',
                            desc='Learning and testing over experiences', position=0, leave=True):

        train_batch_identifiers, current_batch_idf, data_set = combination

        # --- Training Pipeline ---
        if data_set == 'train':
            # For 'from_scratch', train_batch_identifiers will have one ID.
            # For 'cumulative', it will have one or more IDs.
            train_data_list = [data.loc[split_indexes[idf]['train_test']['train_indexes']] for idf in
                               train_batch_identifiers]
            data_batch_train = concat(train_data_list)

            # Initialize lists for final feature names for this training run
            final_input_features = []

            # --- Categorical Feature Processing (Fit and Transform Train Data) ---
            if inputs_categorical_column_names is not None:
                encoder = OneHotEncoder(handle_unknown='ignore', sparse_output=False)
                encoder.fit(data_batch_train[inputs_categorical_column_names])
                encoded_cols = encoder.get_feature_names_out(inputs_categorical_column_names)

                train_encoded_df = DataFrame(
                    encoder.transform(data_batch_train[inputs_categorical_column_names]),
                    columns=encoded_cols,
                    index=data_batch_train.index
                )
                data_batch_train.drop(columns=inputs_categorical_column_names, inplace=True)
                data_batch_train = concat([data_batch_train, train_encoded_df], axis=1)
                final_input_features.extend(encoded_cols)

                # Store the fitted encoder
                trained_encoders[train_batch_identifiers] = encoder

            # --- Numerical Feature Processing (Fit and Transform Train Data) ---
            if inputs_numerical_column_names is not None:
                robust_scaler = RobustScaler()
                data_batch_train[inputs_numerical_column_names] = robust_scaler.fit_transform(
                    data_batch_train[inputs_numerical_column_names])
                final_input_features.extend(inputs_numerical_column_names)

                # Store the fitted scaler
                trained_scalers[train_batch_identifiers] = robust_scaler

            # Extract final training inputs and outputs
            inputs_batch_train = data_batch_train[final_input_features]

            # --- Model Training ---
            if output_regression_column_name is not None:
                outputs_batch_train = data_batch_train[output_regression_column_name]
                model = RandomForestRegressor(n_estimators=number_trees, max_depth=maximum_depth,
                                              random_state=random_seed, n_jobs=-1)
                model.fit(inputs_batch_train, outputs_batch_train)
                trained_models[train_batch_identifiers] = model

            elif output_classification_column_name is not None:
                outputs_batch_train = data_batch_train[output_classification_column_name]
                model = RandomForestClassifier(n_estimators=number_trees, max_depth=maximum_depth,
                                               random_state=random_seed, class_weight='balanced', n_jobs=-1)
                model.fit(inputs_batch_train, outputs_batch_train)
                trained_models[train_batch_identifiers] = model

        # --- Testing (Inference) Pipeline ---
        elif data_set == 'test':
            data_batch_test = data.loc[split_indexes[current_batch_idf]['train_test']['test_indexes']].copy()

            # Retrieve the correct fitted objects
            encoder = trained_encoders.get(train_batch_identifiers)
            scaler = trained_scalers.get(train_batch_identifiers)
            model = trained_models.get(train_batch_identifiers)

            if model is None:
                raise ValueError(f"Model for training batch {train_batch_identifiers} not found.")

            # Initialize list for final feature names
            final_input_features = []

            # --- Categorical Feature Processing (Transform Test Data) ---
            if encoder is not None:
                encoded_cols = encoder.get_feature_names_out(inputs_categorical_column_names)
                test_encoded_df = DataFrame(
                    encoder.transform(data_batch_test[inputs_categorical_column_names]),
                    columns=encoded_cols,
                    index=data_batch_test.index
                )
                data_batch_test.drop(columns=inputs_categorical_column_names, inplace=True)
                data_batch_test = concat([data_batch_test, test_encoded_df], axis=1)
                final_input_features.extend(encoded_cols)

            # --- Numerical Feature Processing (Transform Test Data) ---
            if scaler is not None:
                data_batch_test[inputs_numerical_column_names] = scaler.transform(
                    data_batch_test[inputs_numerical_column_names])
                final_input_features.extend(inputs_numerical_column_names)

            # Extract final test inputs and outputs
            inputs_batch_test = data_batch_test[final_input_features]

            # --- Model Inference ---
            if output_regression_column_name is not None:
                outputs_batch_test = data_batch_test[output_regression_column_name]
                y_pred = model.predict(inputs_batch_test)
                metrics[combination] = _get_regression_metrics(y_true=outputs_batch_test, y_pred=y_pred)

            elif output_classification_column_name is not None:
                outputs_batch_test = data_batch_test[output_classification_column_name]
                probs_hat = model.predict_proba(inputs_batch_test)
                labels_hat = model.predict(inputs_batch_test)

                index2index_map = dict(enumerate(model.classes_))
                index2class_map_batch = {idx: index2class_map[index2index_map[idx]] for idx in index2index_map.keys()}

                metrics_presatur = _get_presaturation_classification_metrics(label_true=outputs_batch_test,
                                                                             label_scores=probs_hat,
                                                                             index2class_map=index2class_map_batch)
                metrics_postsatur = _get_postsaturation_classification_metrics(label_true=outputs_batch_test,
                                                                               label_predicted=labels_hat,
                                                                               index2class_map=index2class_map_batch)
                metrics[combination] = {**metrics_presatur, **metrics_postsatur}

        # Output
    return metrics


# INPUTS CHECKING
def _check_inputs(*, data: DataFrame, inputs_numerical_column_names: Optional[str] = None,
                  inputs_categorical_column_names: Optional[str] = None,
                  output_regression_column_name: Optional[str] = None,
                  output_classification_column_name: Optional[str] = None, date_column_name: Optional[str] = None,
                  period: Optional[str] = None, source_column_name: Optional[str] = None,
                  learning_strategy: Optional[str] = 'from_scratch') -> None:
    """
    Validate the inputs provided for model estimation.

    Parameters
    ----------
    data : DataFrame
        The input data containing features and target variables.

    inputs_numerical_column_names : Optional[str], default=None
        List of column names representing numerical input features, if applicable.

    inputs_categorical_column_names : Optional[str], default=None
        List of column names representing categorical input features, if applicable.

    output_regression_column_name : Optional[str], default=None
        Column name for the regression target variable, if applicable.

    output_classification_column_name : Optional[str], default=None
        Column name for the classification target variable, if applicable.

    date_column_name : Optional[str], default=None
        Column name containing date or time information for temporal batching, if applicable.

    period : Optional[str], default=None
        Period for batching the data ('month' or 'year') when using temporal batching.

    source_column_name : Optional[str], default=None
        Column name representing the source of the data (for multi-source batching).

    learning_strategy : Optional[str], default='from_scratch'
        Defines the learning strategy: 'from_scratch' or 'cumulative'.

    Raises
    ------
    ValueError
        If any input parameters are invalid or inconsistent.
    """

    # Data
    if type(data) is not DataFrame:
        raise TypeError('Data must be encapsulated into a Data frame object.')
    else:
        if data.isnull().values.any():
            raise ValueError('Missing data is present in your data frame object. '
                             'Please, process them before calling this function.')

    # Date column
    if date_column_name is not None:
        if type(date_column_name) is not str:
            raise TypeError('Date column must be specified as a string.')
        if date_column_name not in data.columns:
            raise ValueError('Date column not found in the current data frame.')
        # batching period
        if period is None:
            raise ValueError("A batching period needs to be specified: either 'month' or 'year'.")
        else:
            if period not in ('month', 'year'):
                raise ValueError("Current supported batching periods are 'month' and 'year'.")

    # Source column
    if source_column_name is not None:
        if type(source_column_name) is not str:
            raise TypeError('Source column must be specified as a string.')
        if source_column_name not in data.columns:
            raise ValueError('Source column not found in the current data frame.')

    # Date and source column
    if date_column_name is None and source_column_name is None:
        raise ValueError('Either the date column or the source column needs to the provided.')
    if date_column_name is not None and source_column_name is not None:
        raise ValueError('Just one batching column can be considered (date or source but not both simultaneously).')

    # Inputs numerical columns names
    if inputs_numerical_column_names is not None:
        if type(inputs_numerical_column_names) is not list:
            raise TypeError('Numerical inputs columns need to be encapsulated in a list.')
        else:
            if len(inputs_numerical_column_names) == 0:
                raise ValueError('Numerical inputs column names list is void.')
            else:
                for inp_num_col in inputs_numerical_column_names:
                    if type(inp_num_col) is not str:
                        raise TypeError('Numerical input column must be specified as a string.')
                    else:
                        if inp_num_col not in data.columns:
                            raise ValueError('Numerical input column not found in the current data frame.')

    # Inputs categorical columns names
    if inputs_categorical_column_names is not None:
        if type(inputs_categorical_column_names) is not list:
            raise TypeError('Categorical inputs columns need to be encapsulated in a list.')
        else:
            if len(inputs_categorical_column_names) == 0:
                raise ValueError('Categorical inputs column names list is void.')
            else:
                for inp_cat_col in inputs_categorical_column_names:
                    if type(inp_cat_col) is not str:
                        raise TypeError('Categorical input column must be specified as a string.')
                    else:
                        if inp_cat_col not in data.columns:
                            raise ValueError('Categorical input column not found in the current data frame.')

    # Inputs numerical columns names and inputs categorical columns names
    if inputs_numerical_column_names is None and inputs_categorical_column_names is None:
        raise ValueError('At least one input feature needs to be specified.')

    # Output regression column
    if output_regression_column_name is not None:
        if type(output_regression_column_name) is not str:
            raise TypeError('Regression output column must be specified as a string.')
        if output_regression_column_name not in data.columns:
            raise ValueError('Regression column not found in the current data frame.')

    # Output classification column
    if output_classification_column_name is not None:
        if type(output_classification_column_name) is not str:
            raise TypeError('Classification output column must be specified as a string.')
        if output_classification_column_name not in data.columns:
            raise ValueError('Classification column not found in the current data frame.')

    # Output regression and output classification columns
    if output_regression_column_name is None and output_classification_column_name is None:
        raise ValueError('Either the regression output or the classification output need to the provided.')
    if output_regression_column_name is not None and output_classification_column_name is not None:
        raise ValueError('Just one task can be completed per function call. Leave output_regression or '
                         'output_classification as None.')

    # Learning strategy
    if learning_strategy not in ('from_scratch', 'cumulative'):
        raise ValueError('Unrecognized learning strategy.')
    else:
        if source_column_name is not None and learning_strategy == 'cumulative':
            raise ValueError('Cumulative learning can only be applied to temporal batches.')


# SPLITTING INDEXES OBTAINING
def _generate_split_indexes(*, data: DataFrame, batching_column_name: str) -> dict:
    """
    Generate split indexes based on a specified batching column (e.g., time, source).

    Parameters
    ----------
    data : DataFrame
        The input data containing the features and target variables.

    batching_column_name : str
        The column in the data used for creating splits (e.g., time, source).

    Returns
    -------
    dict:
        A dictionary containing the split indexes for each batch. The keys are batch identifiers,
        and the values are dictionaries with 'train' and 'test' indexes.
    """

    # Splitting settings
    test_ratio = 0.2
    number_folds = 4
    random_seed = 42

    # Memory allocation
    split_indexes_map = dict()

    # Unique identifier values extraction
    identifiers = data[batching_column_name].unique().tolist()

    # Iteration over unique identifiers
    for idf in identifiers:
        if idf in split_indexes_map.keys():
            raise ValueError('Batching value collision.')

        # Memory allocation
        split_indexes_map[idf] = {'train_test': {}, 'puretrain_validation': {}}

        # Data batch extraction
        data_batch = data[data[batching_column_name] == idf]

        # Training and test split
        # train and test sets extraction
        data_batch_train, data_batch_test = train_test_split(
            data_batch, test_size=test_ratio, random_state=random_seed, shuffle=False
        )
        # indexes extraction
        #   training
        indexes_batch_train = data_batch_train.index.to_numpy()
        #   test
        indexes_batch_test = data_batch_test.index.to_numpy()

        # Pure training and validation split
        # initialization
        fold_index = 0
        kfold_splitter = KFold(n_splits=number_folds, random_state=None, shuffle=False)
        # indexes generation
        for puretrain_indexes, validation_indexes in kfold_splitter.split(indexes_batch_train):
            # Arrangement
            # pure training set indexes
            split_indexes_map[idf]['puretrain_validation'][
                (f'kfold_{fold_index}', 'puretrain_indexes')] = puretrain_indexes
            # validation set indexes
            split_indexes_map[idf]['puretrain_validation'][
                (f'kfold_{fold_index}', 'validation_indexes')] = validation_indexes

            # Counter updating
            fold_index += 1

        # Arrangement
        split_indexes_map[idf]['train_test']['train_indexes'] = indexes_batch_train
        split_indexes_map[idf]['train_test']['test_indexes'] = indexes_batch_test

    # Output
    return split_indexes_map


# PERFORMANCE METRICS CALCULATION
# Single-label pre-saturation classification metrics
def _get_presaturation_classification_metrics(*, label_true: ndarray, label_scores: ndarray,
                                              index2class_map: dict) -> dict:
    """
    Calculate classification metrics (before saturation, based on probabilities).

    Parameters
    ----------
    label_true : np.ndarray
        The true class labels.

    label_scores : np.ndarray
        The predicted class probabilities.

    index2class_map : Dict[int, str]
        Mapping from class indices to class labels.

    Returns
    -------
    Dict[str, float]
        A dictionary containing classification metrics based on predicted probabilities.

    """

    # Memory allocation
    metrics = dict()

    # Metrics calculation
    # memory allocation
    auc_classes = []  # area under curve per class

    # Catch warnings
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')

        # single-class
        for index, class_ in index2class_map.items():
            # class identifier generation
            class_idf = str(class_).upper()
            # binarization and extraction of scores per class
            if len(label_true.shape) == 1:
                label_true_class = label_true == index
            else:  # one-hot encoding
                label_true_class = label_true[:, index]
            label_true_class = label_true_class.astype(int)
            label_scores_class = label_scores[:, index]
            # area under curve per class calculation
            try:
                auc_class = skmet.roc_auc_score(label_true_class, label_scores_class)
            except:
                auc_class = 0
                # print('Problem calculating area under curve.')
            # arrangement
            auc_classes.append(auc_class)
            metrics['AUC_' + class_idf] = auc_class

        # multi-class
        # area under curve
        metrics['AUC_MACRO'] = sum(auc_classes) / len(auc_classes)
        # cross-entropy loss
        try:
            metrics['LOGLOSS'] = skmet.log_loss(label_true, label_scores)
        except:
            metrics['LOGLOSS'] = 1
            # print('Problem calculating logloss.')

    # Output
    return metrics


# Single-label post-saturation classification metrics
def _get_postsaturation_classification_metrics(*, label_true: ndarray, label_predicted: ndarray,
                                               index2class_map: dict) -> dict:
    """
    Calculate classification metrics after saturation (i.e., after thresholding the predicted probabilities).

    Parameters
    ----------
    label_true : np.ndarray
        The true class labels.

    label_predicted : np.ndarray
        The predicted class labels after applying a threshold (typically 0.5 for binary classification).

    index2class_map : Dict[int, str]
        Mapping from class indices to class labels.

    Returns
    -------
    Dict[str, float]
        A dictionary containing classification metrics based on predicted labels.

    """

    # Memory allocation
    metrics = dict()

    # Metrics calculation
    # Catch warnings
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')

        # single-class
        for index, class_ in index2class_map.items():
            # class identifier generation
            class_idf = str(class_).upper()
            # binarization
            label_true_binarized = label_true == index
            label_predicted_binarized = label_predicted == index
            # recall
            metrics['RECALL_' + class_idf] = skmet.recall_score(
                label_true_binarized, label_predicted_binarized, average='binary')
            # precision
            metrics['PRECISION_' + class_idf] = skmet.precision_score(
                label_true_binarized, label_predicted_binarized, average='binary')
            # f1_score
            metrics['F1-SCORE_' + class_idf] = skmet.f1_score(
                label_true_binarized, label_predicted_binarized, average='binary')

        # multi-class
        # accuracy
        metrics['ACCURACY'] = skmet.accuracy_score(label_true, label_predicted)
        # recall
        metrics['RECALL_MACRO'] = skmet.recall_score(label_true, label_predicted, average='macro')
        metrics['RECALL_MICRO'] = skmet.recall_score(label_true, label_predicted, average='micro')
        metrics['RECALL_WEIGHTED'] = skmet.recall_score(label_true, label_predicted, average='weighted')
        # precision
        metrics['PRECISION_MACRO'] = skmet.precision_score(label_true, label_predicted, average='macro')
        metrics['PRECISION_MICRO'] = skmet.precision_score(label_true, label_predicted, average='micro')
        metrics['PRECISION_WEIGHTED'] = skmet.precision_score(label_true, label_predicted, average='weighted')
        # f1-score
        metrics['F1-SCORE_MACRO'] = skmet.f1_score(label_true, label_predicted, average='macro')
        metrics['F1-SCORE_MICRO'] = skmet.f1_score(label_true, label_predicted, average='micro')
        metrics['F1-SCORE_WEIGHTED'] = skmet.f1_score(label_true, label_predicted, average='weighted')

    # Output
    return metrics


# Regression metrics
def _get_regression_metrics(*, y_true: ndarray, y_pred: ndarray) -> dict:
    """
    Calculate regression metrics: Mean Absolute Error (MAE), Mean Squared Error (MSE), Root Mean Squared Error (RMSE)
    and R-squared (R2) score.

    Parameters
    ----------
    y_true : DataFrame
        The true target values.

    y_pred : np.ndarray
        The predicted values from the model.

    Returns
    -------
    Dict[str, float]
        A dictionary containing the calculated regression metrics.

    """

    # Memory allocation
    metrics = dict()

    # Metrics calculation
    # mean absolute error
    metrics['MEAN_ABSOLUTE_ERROR'] = skmet.mean_absolute_error(y_true=y_true, y_pred=y_pred)
    # mean squared error
    metrics['MEAN_SQUARED_ERROR'] = skmet.mean_squared_error(y_true=y_true, y_pred=y_pred)
    # root mean squared error
    metrics['ROOT_MEAN_SQUARED_ERROR'] = sqrt(metrics['MEAN_SQUARED_ERROR'])
    # R2
    metrics['R_SQUARED'] = skmet.r2_score(y_true=y_true, y_pred=y_pred)

    # Output
    return metrics



# Copyright 2024 Biomedical Data Science Lab, Universitat Politècnica de València (Spain)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd
from scipy.interpolate import UnivariateSpline

from dashi.unsupervised_characterization.variability_metrics.igt_projection import IGTProjection


def _estimate_igt_trajectory(igt_projection: IGTProjection, number_of_points=None):
    """
    Estimates a trajectory of the information temporal evolution in a IGT projection by fitting a
    cubic smoothing spline
    """
    if igt_projection is None:
        raise ValueError("An input IGT projection object is required.")
    if number_of_points is None:
        number_of_points = igt_projection.projection.shape[0] * 10

    dimensions = igt_projection.projection.shape[1]
    batches = igt_projection.projection.shape[0]

    t = np.arange(1, batches + 1)
    tt = np.linspace(1, batches, num=number_of_points)

    trajectory_function = {}
    points = pd.DataFrame(
        np.zeros((number_of_points, dimensions)),
        columns=[f"D{i + 1}" for i in range(dimensions)]
    )
    dates = pd.date_range(
        start=min(igt_projection.data_temporal_map.dates),
        end=max(igt_projection.data_temporal_map.dates),
        periods=number_of_points
    )

    for i in range(dimensions):
        spline = UnivariateSpline(x=t, y=igt_projection.projection[:, i], s=0.008, k=3)
        trajectory_function[f"D{i + 1}"] = spline
        points.iloc[:, i] = spline(tt)

    results = {"points": points, "dates": dates, "trajectoryFunction": trajectory_function}
    return results



# Copyright 2024 Biomedical Data Science Lab, Universitat Politècnica de València (Spain)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Data Source Map plotting main functions and classes
"""
import numpy as np
import pandas as pd
import plotly.graph_objs as go
import plotly.subplots as sp
import plotly.io as pio
import plotly.colors
from typing import Optional, Dict

from dashi._constants import VALID_SORTING_METHODS, VALID_COLOR_PALETTES, \
    VALID_PLOT_MODES, VALID_STRING_TYPE, VALID_CATEGORICAL_TYPE
from dashi.unsupervised_characterization.data_source_map.data_source_map import DataSourceMap, MultiVariateDataSourceMap
from dashi.unsupervised_characterization.utils import (_validate_plot_args, _sort_support_and_map, _get_counts_array,
                                                       _marginalize_multivariate_map, _create_series_figure)

# pio.renderers.default = "browser"

def plot_univariate_data_source_map(
        data_source_map: DataSourceMap,
        absolute: bool = False,
        log_transform: bool = False,
        start_value: Optional[int] = 0,
        end_value: Optional[int] = None,
        sorting_method: str = 'alphabetical',
        title: Optional[str] = None
) -> go.Figure:
    """
    Plots a Data Source heatmap or series from a DataSourceMap object.

    Parameters
    ----------
    data_source_map : DataTemporalMap
        The DataSourceMap object that contains data to be plotted.

    absolute : bool
        If True, plot absolute values; otherwise, the relative probabilities are plotted. Default is False.

    log_transform : bool
        If True, applies a log transformation to the data for better visibility of small values. Default is False.

    start_value : int, optional
        The value at which to start the plot. Default is 0.

    end_value : int, optional
        The value at which to end the plot. If None, the plot extends to the last value. Default is None.

    sorting_method : str, optional
        The method by which the data will be sorted for display (e.g., 'frequency', 'alphabetical').
        Default is 'frequency'.

    title : str, optional
        The title of the plot. If None, a default title is used. Default is None.

    Returns
    -------
    Figure
        The Plotly figure object representing the plot
    """
    if not type(data_source_map) == DataSourceMap:
        raise TypeError('data_source_map must be an instance of DataSourceMap.')
    _validate_plot_args(
        mode=None,
        color_palette=None,
        absolute=absolute,
        log_transform=log_transform,
        start_value=start_value,
        sorting_method=sorting_method
    )

    if absolute:
        source_map = data_source_map.counts_map
    else:
        source_map = data_source_map.probability_map

    sources = data_source_map.sources
    support = np.array(data_source_map.support.iloc[:, 0].tolist())
    variable_type = data_source_map.variable_type

    support, source_map = _sort_support_and_map(
        support=support,
        data_map=source_map,
        variable_type=variable_type,
        sorting_method=sorting_method
    )

    if not end_value or end_value > source_map.shape[0]:
        end_value = source_map.shape[1]

    if start_value > source_map.shape[0]:
        start_value = source_map.shape[1]

    counts_subarray = _get_counts_array(
        data_map=source_map,
        start_value=start_value,
        end_value=end_value,
        log_transform=log_transform
    )

    font = dict(size=20, color='#7f7f7f')
    x_axis = dict(title=data_source_map.variable_name,
                  titlefont=font,
                  tickvals=support,
                  tickfont={'color': 'black'},
                  ticks='outside',
                  tickcolor='black',
                  tickangle=45
                  )

    figure = _create_series_figure(
        data_map=data_source_map,
        x=support,
        y=counts_subarray,
        name=sources,
        absolute=absolute,
        x_axis=x_axis,
        font=font,
        title=title,
        _range=range(len(data_source_map.sources))
    )
    return figure

def plot_multivariate_data_source_map(
        data_source_map: MultiVariateDataSourceMap,
        absolute: bool = False
) -> go.Figure:
    """
    Plots a multivariate Data Source heatmap from a MultiVariateDataSourceMap object.

    Parameters
    ----------
    data_source_map : MultiVariateDataSourceMap
        The MultiVariateDataSourceMap object that contains multivariate data to be plotted.

    absolute : bool, optional
        If True, plot absolute values; otherwise, the relative probabilities are plotted. Default is False.

    Returns
    -------
    Figure
        The Plotly figure object representing the multivariate heatmap.
    """
    if not type(data_source_map) == MultiVariateDataSourceMap:
        raise TypeError('data_source_map must be an instance of MultiVariateDataSourceMap, obtained from the'
                        ' estimate_multivariate_data_source_map function.')

    if not isinstance(absolute, bool):
        raise TypeError('absolute must be a boolean value, indicating whether to plot absolute counts or probabilities.')

    sources = data_source_map.sources
    supports = data_source_map.multivariate_support
    dimensions = len(supports)

    # Create a color palette for the sources
    palette = plotly.colors.qualitative.Plotly
    colors = {source: palette[i % len(palette)] for i, source in enumerate(sources)}

    if absolute:
        multivariate_map = data_source_map.multivariate_counts_map
    else:
        multivariate_map = data_source_map.multivariate_probability_map

    probability_map_list = _marginalize_multivariate_map(
        multivariate_map=multivariate_map,
        supports=supports ,
        dimensions=dimensions)

    subplot = sp.make_subplots(
        rows=dimensions,
        cols=1,
        shared_xaxes=False,
        vertical_spacing=0.05
    )
    font = dict(size=20, color='#7f7f7f')

    for i, source_map in enumerate(probability_map_list):
        support = np.array(source_map.columns)
        counts_subarray = [row for row in source_map.values]

        for j in range(len(sources)):
            trace = go.Scatter(
                x=support,
                y=counts_subarray[j],
                mode='lines',
                name=str(sources[j]),
                showlegend= (i == 0),
                legendgroup=sources[j],
                line=dict(color=colors[sources[j]])
            )
            subplot.add_trace(trace, row=i + 1, col=1)

        subplot.update_yaxes(
            title=f'PC {i + 1}',
            titlefont=font,
            automargin=True,
            row=i + 1,
            col=1,
            ticks='outside',
            tickcolor='black'
        )

        subplot.update_xaxes(
            tickvals=support,
            tickformat='.2f',
            tickangle=15,
            tickfont={'size': 12},
            title_font=font,
            title_text='Support' if i == dimensions - 1 else None,
            # title_font=font if i == dimensions - 1 else None,
            row=i + 1,
            col=1,
            ticks='outside',
            tickcolor='black'
        )

    subplot.update_layout(
        autosize=True,
        height=min(300 * dimensions, 800),
        showlegend=True,
        legend_title_text='Source',
        template='plotly_white',
        margin=dict(t=60, r=20, b=60, l=60),
        coloraxis=dict(colorscale='Spectral_r'),
        title=f'{"Absolute frequencies" if absolute else "Probability distribution"} '
              f'data source map'
    )

    subplot.show()
    return subplot

def plot_conditional_data_source_map(
        data_source_map_dict: Dict[str, MultiVariateDataSourceMap],
        absolute: bool = False
) -> None:
    """
    Plots a Figure for each dimension selected in the data_temporal_map_dict. Each Figure represents the
    Data Temporal heatmap of each label in that dimension

    Parameters
    ----------
    data_source_map_dict : Dict[str, MultiVariateDataSourceMap]
        A dictionary where keys are labels (strings), and values are the corresponding
        `MultiVariateDataSourceMap` objects obtained from the 'estimate_conditional_data_source_map' function.

    absolute : bool, optional
        If True, plot absolute values; otherwise, relative probabilities are plotted. Default is False.

    Returns
    -------
    None
    """

    if not isinstance(data_source_map_dict, dict) and not all(
            isinstance(value, MultiVariateDataSourceMap) for value in data_source_map_dict.values()
    ):
        raise TypeError('data_source_map_dict must be a dictionary with MultiVariateDataSourceMap instances, resultant'
                        'of the estimate_conditional_data_source_map function.')

    if not isinstance(absolute, bool):
        raise TypeError('absolute must be a boolean value, indicating whether to plot absolute counts or probabilities.')


    labels = list(data_source_map_dict.keys())
    probability_map_dict: dict = {}
    sources_dict: dict = {}

    for label, data_source_map in data_source_map_dict.items():
        sources_dict[label] = data_source_map.sources
        supports = data_source_map.multivariate_support
        dimensions = len(supports)

        if absolute:
            if dimensions == 1:
                multivariate_map = data_source_map.counts_map
            else:
                multivariate_map = data_source_map.multivariate_counts_map
        else:
            if dimensions == 1:
                multivariate_map = data_source_map.probability_map
            else:
                multivariate_map = data_source_map.multivariate_probability_map

        probability_map_list = _marginalize_multivariate_map(
            multivariate_map=multivariate_map,
            supports=supports,
            dimensions=dimensions
        )

        probability_map_dict[label] = probability_map_list

    for dim in range(dimensions):
        subplot = sp.make_subplots(
            rows=len(labels),
            cols=1,
            shared_xaxes=False,
            vertical_spacing=0.06
        )

        font = dict(size=20, color='#7f7f7f')

        for i, (label, probability_map_list) in enumerate(probability_map_dict.items()):
            sources = sources_dict[label]
            source_map = probability_map_list[dim]
            support = np.array(source_map.columns)
            counts_subarray = [row for row in source_map.values]

            # Create a color palette for the sources
            palette = plotly.colors.qualitative.Plotly
            colors = {source: palette[i % len(palette)] for i, source in enumerate(sources)}

            for j in range(len(sources)):
                trace = go.Scatter(
                    x=support,
                    y=counts_subarray[j],
                    mode='lines',
                    name=str(sources[j]),
                    showlegend= (i == 0),
                    legendgroup=sources[j],
                    line=dict(color=colors[sources[j]])
                )
                subplot.add_trace(trace, row=labels.index(label) + 1, col=1)

            subplot.update_yaxes(
                title=f'{label}',
                titlefont=font,
                automargin=True,
                row=labels.index(label) + 1,
                col=1,
                ticks='outside',
                tickcolor='black'
            )

            subplot.update_xaxes(
                tickvals=support,
                tickformat='.2f',
                tickfont={'size': 12},
                tickangle=45,
                title_text='Support' if labels.index(label) == len(labels) - 1 else None,
                title_font=font if labels.index(label) == len(labels) - 1 else None,
                row=labels.index(label) + 1,
                col=1,
                ticks='outside',
                tickcolor='black'
            )

        subplot.update_layout(
            autosize=True,
            height=min(300 * len(labels), 800),
            showlegend=True,
            legend_title_text='Source',
            template='plotly_white',
            margin=dict(t=60, r=20, b=60, l=60),
            coloraxis=dict(colorscale='Spectral_r'),
            title=f'{"Absolute frequencies" if absolute else "Probability distribution"} '
                  f'conditional data source map of Principal Component {dim + 1}'
        )

        subplot.show()
    return None









from .unsupervised_characterization.data_temporal_map.data_temporal_map import *
from .unsupervised_characterization.data_temporal_map.data_temporal_map_plotter import *
from .unsupervised_characterization.variability_metrics.igt_plotting import *
from .unsupervised_characterization.variability_metrics.igt_projection_estimator import *
from .unsupervised_characterization.data_source_map.data_source_map import *
from .unsupervised_characterization.data_source_map.data_source_map_plotter import *
from .unsupervised_characterization.variability_metrics.estimate_msv_metrics import *
from .unsupervised_characterization.variability_metrics.plot_msv import *
from .supervised_characterization.arrange_metrics import *
from .supervised_characterization.estimate_models import *
from .supervised_characterization.plot_performance import *
from .utils import *



# Copyright 2024 Biomedical Data Science Lab, Universitat Politècnica de València (Spain)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Data Temporal Map plotting main functions and classes
"""

from datetime import datetime

import numpy as np
import pandas as pd
import plotly.graph_objs as go
import plotly.subplots as sp
import plotly.io as pio
from typing import Optional, Dict

from dashi._constants import VALID_SORTING_METHODS, VALID_COLOR_PALETTES, \
    VALID_PLOT_MODES, VALID_STRING_TYPE, VALID_CATEGORICAL_TYPE

from dashi.unsupervised_characterization.data_temporal_map.data_temporal_map import (DataTemporalMap,
                                                                                     MultiVariateDataTemporalMap,
                                                                                     trim_data_temporal_map)
from dashi.unsupervised_characterization.utils import (_validate_plot_args, _sort_support_and_map, _get_counts_array,
                                                       _create_heatmap_figure, _create_series_figure,
                                                       _marginalize_multivariate_map)

# SETTINGS
# pio.renderers.default = 'browser'


def plot_univariate_data_temporal_map(
        data_temporal_map: DataTemporalMap,
        absolute: bool = False,
        log_transform: bool = False,
        start_value: Optional[int] = 0,
        end_value: Optional[int] = None,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        sorting_method: str = 'frequency',
        color_palette: str = 'Spectral',
        mode: str = 'heatmap',
        title: Optional[str] = None
) -> go.Figure:
    """
    Plots a Data Temporal heatmap or series from a DataTemporalMap object.

    Parameters
    ----------
    data_temporal_map : DataTemporalMap
        The DataTemporalMap object that contains the temporal data to be plotted.

    absolute : bool
        If True, plot absolute values; otherwise, the relative probabilities are plotted. Default is False.

    log_transform : bool
        If True, applies a log transformation to the data for better visibility of small values. Default is False.

    start_value : int, optional
        The value at which to start the plot. Default is 0.

    end_value : int, optional
        The value at which to end the plot. If None, the plot extends to the last value. Default is None.

    start_date : datetime, optional
        The starting date for the plot (filters the data). If None, uses the first date in the data. Default is None.

    end_date : datetime, optional
        The ending date for the plot (filters the data). If None, uses the last date in the data. Default is None.

    sorting_method : str, optional
        The method by which the data will be sorted for display (e.g., 'frequency', 'alphabetical').
        Default is 'frequency'.

    color_palette : str, optional
        The color palette to be used for the plot (e.g., 'Spectral', 'viridis', 'viridis_r', 'magma', 'magma_r).
        Default is 'Spectral'.

    mode : str, optional
        The mode of visualization (e.g., 'heatmap', 'series'). Default is 'heatmap'.

    title : str, optional
        The title of the plot. If None, a default title is used. Default is None.

    Returns
    -------
    Figure
        The Plotly figure object representing the plot
    """
    if not type(data_temporal_map) == DataTemporalMap:
        raise TypeError('data_temporal_map must be of type DataTemporalMap. For multivariate plot'
                        ' use plot_multivariate_data_temporal_map function')
    _validate_plot_args(
        mode=mode,
        color_palette=color_palette,
        absolute=absolute,
        log_transform=log_transform,
        start_value=start_value,
        sorting_method=sorting_method
    )

    data_temporal_map = trim_data_temporal_map(data_temporal_map, start_date, end_date)

    if absolute:
        temporal_map = data_temporal_map.counts_map
    else:
        temporal_map = data_temporal_map.probability_map

    dates = data_temporal_map.dates

    support = np.array(data_temporal_map.support.iloc[:, 0].tolist())
    variable_type = data_temporal_map.variable_type

    support, temporal_map = _sort_support_and_map(
        support=support,
        data_map=temporal_map,
        variable_type=variable_type,
        sorting_method=sorting_method
    )

    if not end_value or end_value > temporal_map.shape[0]:
        end_value = temporal_map.shape[1]

    if start_value > temporal_map.shape[0]:
        start_value = temporal_map.shape[1]

    counts_subarray = _get_counts_array(
        data_map=temporal_map,
        start_value=start_value,
        end_value=end_value,
        log_transform=log_transform,
        temporal=True
    )

    font = dict(size=20, color='#7f7f7f')
    x_axis = dict(title='Date',
                  tickvals=dates[::2] if len(dates) > 2 else dates,
                  titlefont=font,
                  tickfont={'color': 'black'},
                  type='date',
                  ticks='outside',
                  tickcolor='black')

    if mode == 'heatmap':
        figure = _create_heatmap_figure(
            data_map=data_temporal_map,
            x=dates,
            y=support[start_value:end_value],
            z=counts_subarray,
            color_palette=color_palette,
            font=font,
            x_axis=x_axis,
            title=title,
            absolute=absolute,
        )
        return figure

    elif mode == 'series':
        figure = _create_series_figure(
            data_map=data_temporal_map,
            x=dates,
            y=counts_subarray,
            name=support,
            absolute=absolute,
            x_axis=x_axis,
            font=font,
            title=title,
            _range=range(start_value, end_value),
            temporal=True
        )
        return figure
    return None


def plot_multivariate_data_temporal_map(
        data_temporal_map: MultiVariateDataTemporalMap,
        absolute: bool = False,
        log_transform: bool = False
) -> go.Figure:
    """
    Plots a Data Temporal heatmap from a MultiVariateDataTemporalMap object.

    Parameters
    ----------
    data_temporal_map : MultiVariateDataTemporalMap
        The MultiVariateDataTemporalMap object that contains the temporal data to be plotted.

    absolute : bool, optional
        If True, plot absolute values; otherwise, the relative probabilities are plotted. Default is False.

    log_transform : bool
        If True, applies a log transformation to the data for better visibility of small values. Default is False.


    Returns
    -------
    Figure
        The Plotly figure object representing the plot.
    """
    if not type(data_temporal_map) == MultiVariateDataTemporalMap:
        raise TypeError('data_temporal_map must be of type MultiVariateDataTemporalMap, obtained from the '
                        'estimate_multivariate_data_temporal_map function.')

    if not isinstance(absolute, bool):
        raise TypeError('absolute must be a boolean value, indicating whether to plot absolute counts or probabilities.')

    dates = data_temporal_map.dates

    supports = data_temporal_map.multivariate_support
    dimensions = len(supports)

    if absolute:
        multivariate_map = data_temporal_map.multivariate_counts_map
    else:
        multivariate_map = data_temporal_map.multivariate_probability_map

    probability_map_list = _marginalize_multivariate_map(
        multivariate_map=multivariate_map,
        supports=supports,
        dimensions=dimensions)

    subplot = sp.make_subplots(rows=dimensions,
                               cols=1,
                               shared_xaxes=True,
                               vertical_spacing=0.02
                               )

    font = dict(size=20, color='#7f7f7f')
    x_axis_tickvals = dates[::2] if len(dates) > 2 else dates

    for i, temporal_map in enumerate(probability_map_list):
        support = np.array(temporal_map.columns)
        if log_transform:
            temporal_map = np.log(temporal_map + 1e-8)
        counts_subarray = [row for row in temporal_map.values]
        counts_subarray = list(zip(*counts_subarray))

        figure = go.Heatmap(
            x=dates,
            y=support,
            z=counts_subarray,
            reversescale=True,
            coloraxis='coloraxis'
        )

        subplot.add_trace(figure, row=i + 1, col=1)

        subplot.update_yaxes(
            title=f'PC {i + 1}',
            titlefont=font,
            automargin=True,
            row=i + 1,
            col=1,
            ticks='outside',
            tickcolor='black'
        )

        subplot.update_xaxes(
            tickvals=x_axis_tickvals,
            tickfont={'size': 12},
            type='date',
            title_text='Date' if i == dimensions - 1 else None,
            title_font=font if i == dimensions - 1 else None,
            row=i + 1,
            col=1,
            ticks='outside',
            tickcolor='black'
        )

    subplot.update_layout(
        autosize=True,
        height=min(300 * dimensions, 800),
        showlegend=False,
        template='plotly_white',
        margin=dict(t=60, r=20, b=60, l=60),
        coloraxis=dict(colorscale='Spectral_r'),
        title=f'{"Absolute frequencies" if absolute else "Probability distribution"} '
              f'data temporal heatmap'
    )

    subplot.show()
    return subplot


def plot_conditional_data_temporal_map(
        data_temporal_map_dict: Dict[str, MultiVariateDataTemporalMap],
        absolute: bool = False,
        log_transform: bool = False
):
    """
    Plots a Figure for each dimension selected in the data_temporal_map_dict. Each Figure represents the
    Data Temporal heatmap of each label in that dimension

    Parameters
    ----------
    data_temporal_map_dict : Dict[str, MultiVariateDataTemporalMap]
        A dictionary where keys are labels (strings), and values are the corresponding
        `MultiVariateDataTemporalMap` objects obtained from the 'estimate_conditional_data_temporal_map' function.

    absolute : bool, optional
        If True, plot absolute values; otherwise, relative probabilities are plotted. Default is False.

    log_transform : bool
        If True, applies a log transformation to the data for better visibility of small values. Default is False.

    Returns
    -------
    None
    """
    if not type(data_temporal_map_dict) == dict:
        raise TypeError('data_temporal_map must be a dictionary of objects MultiVariateDataTemporalMap, resultant of '
                        'the estimate_conditional_data_temporal_map function')

    if not isinstance(absolute, bool):
        raise ValueError('absolute must be a boolean value, indicating whether to plot absolute counts or probabilities.')

    labels = list(data_temporal_map_dict.keys())
    probability_map_dict = dict()
    dates_dict = dict()
    for label, data_temporal_map in data_temporal_map_dict.items():
        dates_dict[label] = data_temporal_map.dates
        supports = data_temporal_map.multivariate_support
        dimensions = len(supports)

        if absolute:
            if dimensions == 1:
                multivariate_map = data_temporal_map.counts_map
            else:
                multivariate_map = data_temporal_map.multivariate_counts_map
        else:
            if dimensions == 1:
                multivariate_map = data_temporal_map.probability_map
            else:
                multivariate_map = data_temporal_map.multivariate_probability_map

        probability_map_list = _marginalize_multivariate_map(
            multivariate_map=multivariate_map,
            supports=supports,
            dimensions=dimensions)

        probability_map_dict[label] = probability_map_list

    for dim in range(dimensions):
        subplot = sp.make_subplots(rows=len(labels),
                                   cols=1,
                                   shared_xaxes=True,
                                   vertical_spacing=0.04
                                   )

        font = dict(size=20, color='#7f7f7f')

        for label, probability_map_list in probability_map_dict.items():
            dates = dates_dict[label]
            x_axis_tickvals = dates[::2] if len(dates) > 2 else dates
            temporal_map = probability_map_list[dim]
            support = np.array(temporal_map.columns)
            if log_transform:
                temporal_map = np.log(temporal_map + 1e-8)
            counts_subarray = [row for row in temporal_map.values]
            counts_subarray = list(zip(*counts_subarray))

            figure = go.Heatmap(
                x=dates.astype(str),
                y=support,
                z=counts_subarray,
                reversescale=True,
                coloraxis='coloraxis'
            )

            subplot.add_trace(figure, row=labels.index(label) + 1, col=1)

            subplot.update_yaxes(
                title=f'{label}',
                titlefont=font,
                automargin=True,
                row=labels.index(label) + 1,
                col=1,
                ticks='outside',
                tickcolor='black'
            )

            subplot.update_xaxes(
                tickvals=x_axis_tickvals,
                tickfont={'size': 12},
                type='date',
                title_text='Date' if labels.index(label) == len(labels) - 1 else None,
                title_font=font if labels.index(label) == len(labels) - 1 else None,
                row=labels.index(label) + 1,
                col=1,
                ticks='outside',
                tickcolor='black'
            )

        subplot.update_layout(
            autosize=True,
            height=min(300 * len(labels), 800),
            showlegend=False,
            template='plotly_white',
            margin=dict(t=60, r=20, b=60, l=60),
            coloraxis=dict(colorscale='Spectral_r'),
            title=f'{"Absolute frequencies" if absolute else "Probability distribution"} '
                  f'conditional data temporal heatmap of Principal Component {dim + 1}'
        )

        subplot.show()
    return None



# Copyright 2024 Biomedical Data Science Lab, Universitat Politècnica de València (Spain)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utils functions
"""

from datetime import datetime
from typing import Optional, List

import pandas as pd

from dashi._constants import MONTH_SHORT_ABBREVIATIONS, VALID_DATE_TYPE


def _format_date_for_year(date: datetime) -> str:
    year_part = date.strftime('%Y')

    return year_part


def _format_date_for_month(date: datetime) -> str:
    year_part = date.strftime('%y')
    month_part = MONTH_SHORT_ABBREVIATIONS[date.month - 1]

    return year_part + month_part


def _format_date_for_week(date: datetime) -> str:
    year_part = date.strftime('%y')
    month_part = MONTH_SHORT_ABBREVIATIONS[date.month - 1]
    day_part = str(date.isoweekday())

    return year_part + month_part + day_part

def _validate_date_format(date_format, verbose=False):
    """Valida el formato de fecha y muestra mensajes informativos si verbose=True."""
    has_year = _is_letter_in_date_format(date_format, ['Y', 'y'])
    has_month = _is_letter_in_date_format(date_format, ['m', 'M', 'b', 'B', 'h'])
    has_day = _is_letter_in_date_format(date_format, ['d', 'D'])

    if has_year and has_month and has_day:
        if verbose:
            print('The data format contains year, month, and day')
    elif has_year and has_month:
        if verbose:
            print('The data format contains year and month but not day')
            print('Take into account that if you perform an analysis by week, the day will be automatically assigned as the first day of the month.')
    elif has_year:
        if verbose:
            print('The data format contains only the year')
            print('Take into account that if you perform an analysis by week or by month, they will be automatically assigned as the first day of the month and first month of the year.')
    else:
        print('Please, check the format of the date. At least it should contain the year.')
        raise ValueError('Invalid date format')


def format_data(input_dataframe: pd.DataFrame,
                *,
                date_column_name: Optional[str] = None,
                source_column_name: Optional[str] = None,
                date_format: Optional[str] = '%y/%m/%d',
                verbose: bool = False,
                numerical_column_names: Optional[List[str]] = None,
                categorical_column_names: Optional[List[str]] = None) -> pd.DataFrame:
    """
    Function to transform dates into 'Date' Python format

    Parameters
    ----------
    input_dataframe : pd.DataFrame
        Pandas dataframe object with at least one columns of dates.

    date_column_name: Optional[str]
        The name of the column containing the dates. If you are not performing a temporal analysis, set this parameter
        to `None`.

    source_column_name: Optional[str]
        The name of the column containing the source information. If you are not performing a multi-source analysis,
        set this parameter to `None`.

    date_format: Optional[str]
        Structure of date format. By default '%y/%m/%d'.

    verbose: bool
        Whether to display additional information during the process. Defaults to `False`.

    numerical_column_names: Optional[List[str]]
        A list containing all the numerical column names in the dataset. If this parameter is `None`, the variables
        types must be managed by the user.

    categorical_column_names: Optional[List[str]]
        A list containing all the categorical column names in the dataset. If this parameter is `None`, the variables
        types must be managed by the user.

    Returns
    -------
    pd.DataFrame
        A `pandas.DataFrame` with each column cast to its correct dtype and any rows containing missing values in the
        date or source fields dropped.
    """
    if date_column_name is not None and date_column_name not in input_dataframe.columns:
        raise ValueError(f'There is no column in your DataFrame named as: {date_column_name}')

    if source_column_name is not None and source_column_name not in input_dataframe.columns:
        raise ValueError(f'There is no column in your DataFrame named as: {source_column_name}')

    output_dataframe = input_dataframe.copy()

    if numerical_column_names is not None:
        if verbose:
            print('Formating numerical columns as float')
        for col in numerical_column_names:
            output_dataframe[col] = output_dataframe[col].astype(float)

    if categorical_column_names is not None:
        if verbose:
            print('Formating categorical columns as category')
        for col in categorical_column_names:
            output_dataframe[col] = output_dataframe[col].astype(str).astype('category')


    if date_column_name is not None:
        if output_dataframe[date_column_name].dtype == VALID_DATE_TYPE:
            return output_dataframe

        if verbose:
            print(f'Formatting the {date_column_name} column')

        _validate_date_format(date_format, verbose)

        output_dataframe[date_column_name] = pd.to_datetime(output_dataframe[date_column_name], format=date_format)

        initial_len = len(output_dataframe)
        output_dataframe = output_dataframe.dropna(subset=[date_column_name]).reset_index(drop=True)
        removed_rows = initial_len - len(output_dataframe)
        if removed_rows > 0:
            print(f'There are {removed_rows} rows that do not contain date information. They have been removed.')
        return output_dataframe

    elif source_column_name is not None:
        initial_len = len(output_dataframe)
        output_dataframe = output_dataframe.dropna(subset=[source_column_name]).reset_index(drop=True)
        removed_rows = initial_len - len(output_dataframe)
        if removed_rows > 0:
            print(f'There are {removed_rows} rows that do not contain source information. They have been removed.')
        return output_dataframe

    else:
        raise ValueError('Please specify the date column or source column.')


def _is_letter_in_date_format(date_format, date_pattern):
    # Check if any of the pattern elements is on the date_format string
    return any(any(element in character_to_check for element in date_pattern) for character_to_check in date_format)



# Copyright 2024 Biomedical Data Science Lab, Universitat Politècnica de València (Spain)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Information Geometric Temporal class creation
"""

from dataclasses import dataclass
from typing import Union, List

from dashi.unsupervised_characterization.data_temporal_map.data_temporal_map import DataTemporalMap, MultiVariateDataTemporalMap


@dataclass
class IGTProjection:
    """
    Class IGTProjection object contains the estimated Information Geometric Temporal projection
    of a DataTemporalMap or MultiVariateDataTemporalMap objects

    Attributes
    ----------
    data_temporal_map : Union[DataTemporalMap, MultiVariateDataTemporalMap, None]
        A `DataTemporalMap` or 'MultiVariateDataTemporalMap' object containing the temporal data map
        that was used for generating the projection.

    projection : Union[List[List[float]], None]
        A 2D matrix matrix of floats representing the lower-dimensional projection of the temporal data.
        Each row corresponds to a data timestamp, with each column representing a dimension of the projection.

    embedding_type : Union[str, None]
        A string representing the type of embedding used for the projection (e.g., "classicalmds", "nonmetricmds",
        "pca").

    stress : Union[float, None]
        A float value representing the stress (or error) of the projection if it is available (used in MDS
        or similar techniques). The lower the stress value, the better the projection reflects the original data.
        This attribute is `None` if stress is not computed or available.
    """
    data_temporal_map: Union[DataTemporalMap, MultiVariateDataTemporalMap, None] = None
    projection: Union[List[List[float]], None] = None
    embedding_type: Union[str, None] = None
    stress: Union[float, None] = None




import numpy as np
import pandas as pd
from scipy.stats import gaussian_kde
from scipy.linalg import eigh
from dashi._constants import (VALID_FLOAT_TYPE, VALID_CATEGORICAL_TYPE, VALID_INTEGER_TYPE, VALID_STRING_TYPE,
                              VALID_DATE_TYPE, VALID_CONVERSION_STRING_TYPE, VALID_PLOT_MODES, VALID_COLOR_PALETTES,
                              VALID_SORTING_METHODS, MISSING_VALUE)
import plotly.graph_objects as go
import plotly.subplots as sp
import plotly.io as pio
from dataclasses import dataclass
from typing import Optional, List, Union
import warnings
import prince
import plotly.express as px
import plotly.io as pio
from plotly.colors import sample_colorscale

# pio.renderers.default = "browser"


def _estimate_absolute_frequencies(data, varclass, support, numeric_smoothing=False):
    """
    Estimates the absolute frequencies of data, which will be the counts_map in the final DataTemporalMap object
    """
    data = np.array(data)
    if varclass == VALID_STRING_TYPE:
        value_counts = pd.Series(data).value_counts()
        map_data = value_counts.reindex(support, fill_value=0).values

    elif varclass == VALID_FLOAT_TYPE:
        if np.all(np.isnan(data)):
            map_data = np.array([np.nan] * len(support))
        else:
            if not numeric_smoothing:
                hist_support = np.append(support, support[-1] + (support[-1] - support[-2]))
                data = data[(data >= min(hist_support)) & (data < max(hist_support))]
                bin_edges = hist_support
                map_data, _ = np.histogram(data, bins=bin_edges)
            else:
                if np.sum(~np.isnan(data)) < 4:
                    print(
                        'Estimating a 1-dimensional kernel density smoothing with less than 4 data points can result in an inaccurate estimation.'
                        ' For more information see "Density Estimation for Statistics and Data Analysis, Bernard.W.Silverman, CRC, 1986", chapter 4.5.2 "Required sample size for given accuracy".'
                    )
                if np.sum(~np.isnan(data)) < 2:
                    data = np.repeat(data[~np.isnan(data)], 2)
                    ndata = 1
                else:
                    data = data[~np.isnan(data)]
                    ndata = np.sum(~np.isnan(data))

                kde = gaussian_kde(
                    data)
                map_data = kde(support) * ndata

    elif varclass == VALID_INTEGER_TYPE:
        if np.all(np.isnan(data)):
            map_data = np.array([np.nan] * len(support))
        else:
            if all(isinstance(item, pd.Timestamp) for item in support):
                support_int = [item.value for item in support]
                support = support_int
            hist_support = np.append(support, support[-1] + (support[-1] - support[-2]))
            data = data[(data >= min(hist_support)) & (data < max(hist_support))]
            bin_edges = hist_support
            map_data, _ = np.histogram(data, bins=bin_edges)

    else:
        raise ValueError(f'data class {varclass} not valid for distribution estimation.')

    return map_data

def _create_supports(data, supports, columns_types, number_of_columns, numeric_variables_bins, verbose):
    supports_to_fill = {column: None for column in data.columns}
    supports_to_estimate_columns = data.columns.to_series()

    supports_to_fill = {column: None for column in data.columns}
    supports_to_estimate_columns = data.columns.to_series()

    if supports is not None:
        for column_index, column in enumerate(supports):
            if column in supports_to_fill:
                supports_to_fill[column] = supports[column]
                supports_to_estimate_columns.drop(column)
                error_in_support = False

                if supports[column].dtypes == VALID_CATEGORICAL_TYPE:
                    error_in_support = (
                            not supports[column].dtype.name == VALID_CATEGORICAL_TYPE
                            or not supports[column].dtype.name == VALID_STRING_TYPE
                    )
                elif supports[column].dtypes == VALID_DATE_TYPE:
                    error_in_support = not supports[column].dtype.name == VALID_DATE_TYPE
                elif supports[column].dtypes == VALID_INTEGER_TYPE:
                    error_in_support = not supports[column].dtype.name == VALID_INTEGER_TYPE
                elif supports[column].dtypes == VALID_FLOAT_TYPE:
                    error_in_support = not supports[column].dtype.name == VALID_FLOAT_TYPE

                if error_in_support:
                    raise ValueError(
                        f'The provided support for variable {column} does not match with its variable type')

    supports = supports_to_fill

    if any(supports_to_estimate_columns):
        if verbose:
            print('Estimating supports from data')

        all_na = data.loc[:, supports_to_estimate_columns].apply(lambda x: x.isnull().all())

        # Exclude from the analysis those variables with no finite values, if any
        if any(all_na):
            if verbose:
                print(
                    f'Removing variables with no finite values: {", ".join(data.columns[all_na])}')
            warnings.warn(
                f'Removing variables with no finite values: {", ".join(data.columns[all_na])}')

            data = data.loc[:, ~all_na]
            number_of_columns = len(data.columns)
            supports = {column_name: data_type for column_name, data_type in supports.items() if
                        not all_na[column_name]}

            data_types, columns_types = _get_types(data, verbose=False)

    mask = columns_types['categorical'] & supports_to_estimate_columns
    if mask.any():
        for name, col in data.loc[:, mask].items():
            if col.isna().any():
                # Add the category only if it's not already there
                if MISSING_VALUE not in col.cat.categories:
                    col = col.cat.add_categories([MISSING_VALUE])
                # Fill NaNs with the new category
                data[name] = col.fillna(MISSING_VALUE)

        # Extract levels and assign them to supports
        selected_columns = data.loc[:, columns_types['categorical'] & supports_to_estimate_columns]
        levels = selected_columns.apply(lambda col: col.cat.categories)
        supports.update(
            {
                column: levels[column]
                for column
                in data.columns[columns_types['categorical'] & supports_to_estimate_columns]
            }
        )

    if np.any(columns_types['float'] & supports_to_estimate_columns):
        minimums = data.loc[:, columns_types['float'] & supports_to_estimate_columns].apply(np.nanmin,
                                                                                                   axis=0)
        maximums = data.loc[:, columns_types['float'] & supports_to_estimate_columns].apply(np.nanmax,
                                                                                                   axis=0)
        supports.update(
            {
                column: np.linspace(minimum, maximum, numeric_variables_bins).tolist()
                for column, minimum, maximum
                in zip(data.columns[columns_types['float'] & supports_to_estimate_columns], minimums,
                       maximums)
            }
        )
        if np.any(minimums == maximums):
            mask = (minimums == maximums) & columns_types['float'] & supports_to_estimate_columns
            supports.update(
                {
                    column: [value[0] for value in supports[column]]
                    for column
                    in data.columns[mask]
                }
            )

    if np.any(columns_types['integer'] & supports_to_estimate_columns):
        minimums = data.loc[:, columns_types['integer'] & supports_to_estimate_columns].apply(np.nanmin,
                                                                                                     axis=0)
        maximums = data.loc[:, columns_types['integer'] & supports_to_estimate_columns].apply(np.nanmax,
                                                                                                     axis=0)
        if np.sum(columns_types['integer'] & supports_to_estimate_columns) == 1:
            supports.update(
                {
                    column: np.linspace(minimum, maximum, numeric_variables_bins).tolist()
                    for column, minimum, maximum
                    in
                    zip(data.columns[columns_types['integer'] & supports_to_estimate_columns], minimums,
                        maximums)
                }
            )
        else:
            supports.update(
                {
                    column: np.linspace(minimum, maximum, numeric_variables_bins).tolist()
                    for column, minimum, maximum
                    in
                    zip(data.columns[columns_types['integer'] & supports_to_estimate_columns], minimums,
                        maximums)
                }
            )

    if np.any(columns_types['string'] & supports_to_estimate_columns):
        supports.update(
            {
                column: data[column].unique().tolist()
                for column
                in data.columns[columns_types['string'] & supports_to_estimate_columns]
            }
        )

    if np.any(columns_types['date'] & supports_to_estimate_columns):
        minimums = data.loc[:, columns_types['date'] & supports_to_estimate_columns].apply(np.nanmin,
                                                                                                  axis=0)
        maximums = data.loc[:, columns_types['date'] & supports_to_estimate_columns].apply(np.nanmax,
                                                                                                  axis=0)
        supports.update(
            {
                column: pd.date_range(minimum, maximum, periods=numeric_variables_bins).tolist()
                for column, minimum, maximum
                in zip(data.columns[columns_types['date'] & supports_to_estimate_columns], minimums,
                       maximums)
            }
        )

    # Convert factor variables to characters, as used by the xts Objects
    if np.any(columns_types['categorical']):
        converted_columns = data.loc[:, columns_types['categorical']].astype(
            VALID_CONVERSION_STRING_TYPE)
        data = data.assign(**converted_columns)

    # Exclude from the analysis those variables with a single value, if any
    support_lengths = [len(supports[column]) for column in data.columns]
    support_singles_indexes = np.array(support_lengths) < 2
    if np.any(support_singles_indexes):
        if verbose:
            print(
                f'Removing variables with less than two distinct values in their supports: {", ".join(data.columns[support_singles_indexes])}')
        print(
            f'The following variable/s have less than two distinct values in their supports and were excluded from the analysis: {", ".join(data.columns[support_singles_indexes])}')
        data = data.loc[:, ~support_singles_indexes]
        supports = {
            column: supports[column]
            for column
            in data.columns
        }
        number_of_columns = len(data.columns)

    if number_of_columns == 0:
        raise ValueError('Zero remaining variables to be analyzed.')

    return data, supports

def _get_types(data, verbose=False):
    data_types = data.dtypes

    float_columns = data_types == VALID_FLOAT_TYPE
    integer_columns = data_types == VALID_INTEGER_TYPE
    string_columns = data_types == VALID_STRING_TYPE
    date_columns = data_types == VALID_DATE_TYPE
    categorical_columns = data_types == VALID_CATEGORICAL_TYPE

    columns_by_type = {
        "float": float_columns,
        "integer": integer_columns,
        "string": string_columns,
        "date": date_columns,
        "categorical": categorical_columns
    }

    if verbose:
        if any(columns_by_type["float"]):
            print(f'Number of float columns: {sum(columns_by_type["float"])}')
        if any(columns_by_type["integer"]):
            print(f'Number of integer columns: {sum(columns_by_type["integer"])}')
        if any(columns_by_type["string"]):
            print(f'Number of string columns: {sum(columns_by_type["string"])}')
        if any(columns_by_type["date"]):
            print(f'Number of date columns: {sum(columns_by_type["date"])}')
        if any(columns_by_type["categorical"]):
            print(f'Number of categorical columns: {sum(columns_by_type["categorical"])}')

    return data_types, columns_by_type

def _validate_plot_args(
        mode,
        color_palette,
        absolute,
        log_transform,
        start_value,
        sorting_method,
) -> None:
    if mode is not None and mode not in VALID_PLOT_MODES:
        raise ValueError(f'mode must be one of the defined in {VALID_PLOT_MODES}')

    if color_palette is not None and color_palette not in VALID_COLOR_PALETTES:
        raise ValueError(f'color_palette must be one of the defined in {VALID_COLOR_PALETTES}')

    if not isinstance(absolute, bool):
        raise ValueError('absolute must be a logical value')

    if not isinstance(log_transform, bool):
        raise ValueError('log_transform must be a logical value')

    if not isinstance(start_value, int) and start_value < 1:
        raise ValueError('start_value must be greater or equal than 1')

    if sorting_method not in VALID_SORTING_METHODS:
        raise ValueError(f'sorting_method must be one of the defined in {VALID_SORTING_METHODS}')

def _sort_support_and_map(
        support,
        data_map,
        variable_type,
        sorting_method
):
    if variable_type in [VALID_STRING_TYPE, VALID_CATEGORICAL_TYPE]:
        if sorting_method == 'frequency':
            support_order = np.argsort(np.sum(data_map, axis=0))[::-1]
        else:
            support_order = np.argsort(support)

        support = support[support_order]

        # Resort temporal map by support_order
        data_map = np.array([row[support_order] for row in data_map])

        any_supp_na = pd.isnull(support)
        if any_supp_na.any():
            support[any_supp_na] = '<NA>'

    return support, data_map

def _get_counts_array(
        data_map,
        start_value,
        end_value,
        log_transform,
        temporal=False
):
    if log_transform:
        data_map = np.log(data_map + 1e-8)

    counts_subarray = [row[start_value: end_value] for row in data_map]
    if temporal:
        counts_subarray = list(zip(*counts_subarray))

    return counts_subarray

def _create_heatmap_figure(
        data_map,
        x,
        y,
        z,
        color_palette,
        font,
        x_axis,
        title,
        absolute
):
    figure = go.Figure(
        data=go.Heatmap(
            x=x,
            y=y,
            z=z,
            colorscale=color_palette,
            reversescale=True,
            colorbar=dict(
                tickfont=dict(
                    color='black'
                )
            )
        )
    )

    y_axis = dict(
        title=data_map.variable_name,
        titlefont=font,
        automargin=True,
        tickfont={'color': 'black'},
        ticks='outside',
        tickcolor='black'
    )

    figure.update_xaxes(x_axis)

    figure.update_layout(yaxis=y_axis,
                         autosize=True,
                         paper_bgcolor='white',
                         plot_bgcolor='white'
                         )

    if data_map.variable_type in [VALID_STRING_TYPE, VALID_CATEGORICAL_TYPE]:
        figure.update_layout(yaxis_type='category')

    if title is not None:
        figure.update_layout(title={'text': title,
                                    'font': {'color': 'black'
                                             }
                                    })
    else:
        title = 'Probability distribution data temporal heatmap'
        if absolute:
            title = 'Absolute frequencies data temporal heatmap'
        figure.update_layout(title={'text': title,
                                    'font': {'color': 'black'
                                             }
                                    }
                             )
    figure.show()

    return figure

def _create_series_figure(
        data_map,
        x,
        y,
        name,
        absolute,
        x_axis,
        font,
        title,
        _range,
        **kwargs
):
    figure = go.Figure()

    for i in _range:
        trace = go.Scatter(
            x=x,
            y=y[i],
            mode='lines',
            name=str(name[i])
        )
        figure.add_trace(trace)

    if absolute:
        y_axis_title = 'Absolute frequency'
    else:
        y_axis_title = 'Relative frequency'

    y_axis = dict(
        title=y_axis_title,
        titlefont=font,
        automargin=True,
        tickfont={'color': 'black'},
        ticks='outside',
        tickcolor='black'
    )

    figure.update_xaxes(
        x_axis
    )

    figure.update_layout(
        autosize=True,
        yaxis=y_axis,
        template='plotly_white',
        paper_bgcolor='white',
        plot_bgcolor='white',
        legend_title_text=data_map.variable_name if 'temporal' in kwargs else 'Source'
    )

    if title is not None:
        figure.update_layout(title={'text': title,
                                    'font': {'color': 'black'
                                             }
                                    }
                             )
    else:
        figure.update_layout(title={'text': 'Evolution of ' + data_map.variable_name,
                                    'font': {'color': 'black'
                                             }
                                    }
                             )
    figure.show()

    return figure

def _js_divergence(p, q, epsilon=1e-10):
    """
    Computes the Jensen-Shannon (JS) divergence between two probability distributions.

    The Jensen-Shannon divergence is a symmetric and smoothed version of the Kullback-Leibler (KL) divergence
    and measures the similarity between two probability distributions. Unlike the KL divergence, which is asymmetric,
    the JS divergence is symmetric and bounded, making it a more stable measure for comparing distributions.

    The JS divergence is calculated as:
    JS(p || q) = 0.5 * (KL(p || m) + KL(q || m))
    where m = 0.5 * (p + q) is the average distribution, and KL(p || m) is the Kullback-Leibler divergence between
    distribution `p` and `m`.
    """
    p = np.asarray(p)
    q = np.asarray(q)

    p = np.where(p < epsilon, epsilon, p)
    q = np.where(q < epsilon, epsilon, q)

    m = 0.5 * (p + q)

    kl_p_m = np.where(p != 0, p * np.log2(p / m), 0)
    kl_q_m = np.where(q != 0, q * np.log2(q / m), 0)

    result = 0.5 * (np.nansum(kl_p_m) + np.nansum(kl_q_m))

    return result

def _cmdscale(d, k=2, eig=False, add=False, x_ret=False):
    """
    Performs Classical Multidimensional Scaling (MDS) on a distance matrix to reduce the dimensionality
    of the data, while preserving pairwise distances as much as possible.
    """
    # Check for NA values (Not Applicable in numpy, but we can check for NaN)
    if np.isnan(d).any():
        raise ValueError("NA values not allowed in 'd'")

    list_ = eig or add or x_ret

    if not list_:
        if eig:
            print("Warning: eig=TRUE is disregarded when list_=FALSE")
        if x_ret:
            print("Warning: x_ret=TRUE is disregarded when list_=FALSE")

    if not isinstance(d, np.ndarray) or len(d.shape) != 2 or d.shape[0] != d.shape[1]:
        if add:
            d = np.array(d)
        x = np.array(d ** 2, dtype=np.double)
        n = x.shape[0]
        if n != x.shape[1]:
            raise ValueError("distances must be result of 'dist' or a square matrix")
        rn = np.arange(n)
    else:
        n = d.shape[0]
        rn = np.arange(n)
        x = np.zeros((n, n))
        if add:
            d0 = x.copy()
        triu_indices = np.triu_indices_from(x, 1)
        x[triu_indices] = d[triu_indices] ** 2
        x += x.T
        if add:
            d0[triu_indices] = d[triu_indices]
            d = d0 + d0.T

    if not isinstance(n, int) or n > 46340:
        raise ValueError("invalid value of 'n'")

    if k > n - 1 or k < 1:
        raise ValueError("'k' must be in {1, 2, ..  n - 1}")

    # Double centering
    H = np.eye(n) - np.ones((n, n)) / n
    B = -0.5 * H.dot(x).dot(H)

    if add:
        i2 = n + np.arange(n)
        Z = np.zeros((2 * n, 2 * n))
        Z[np.arange(n), i2] = -1
        Z[i2, np.arange(n)] = -x
        Z[i2, i2] = 2 * d
        e = np.linalg.eigvals(Z)
        add_c = np.max(np.real(e))
        x = np.zeros((n, n), dtype=np.double)
        non_diag = np.triu_indices_from(d, 1)
        x[non_diag] = (d[non_diag] + add_c) ** 2
        x = -0.5 * H.dot(x).dot(H)

    e_vals, e_vecs = eigh(B)
    idx = np.argsort(e_vals)[::-1]
    e_vals = e_vals[idx]
    e_vecs = e_vecs[:, idx]

    ev = e_vals[:k]
    evec = e_vecs[:, :k]
    k1 = np.sum(ev > 0)

    if k1 < k:
        print(f"Warning: only {k1} of the first {k} eigenvalues are > 0")
        evec = evec[:, ev > 0]
        ev = ev[ev > 0]

    points = evec * np.sqrt(ev)

    if list_:
        evalus = e_vals
        return {
            'points': points,
            'eig': evalus if eig else None,
            'x': B if x_ret else None,
            'ac': add_c if add else 0,
            'GOF': np.sum(ev) / np.array([np.sum(np.abs(evalus)), np.sum(np.maximum(evalus, 0))])
        }
    else:
        return points

@dataclass
class BaseMultiVariateMap:
    multivariate_probability_map: Optional[np.array] = None
    multivariate_counts_map: Optional[np.array] = None
    multivariate_support: Optional[np.array] = None

    def check_multivariate(self) -> Union[List[str], bool]:
        errors = []
        if self.multivariate_probability_map is not None and self.multivariate_counts_map is not None:
            if len(self.multivariate_probability_map) != len(self.multivariate_counts_map) or \
               any(len(prob_row) != len(count_row)
                   for prob_row, count_row in zip(self.multivariate_probability_map, self.multivariate_counts_map)):
                errors.append("The dimensions of multivariate_probability_map and multivariate_counts_map do not match.")
        if self.multivariate_support is not None and self.multivariate_probability_map is not None:
            if len(self.multivariate_support) != len(self.multivariate_probability_map[0]):
                errors.append("The length of multivariate_support must match the columns of multivariate_probability_map.")
        if self.multivariate_support is not None and self.multivariate_counts_map is not None:
            if len(self.multivariate_support) != len(self.multivariate_counts_map[0]):
                errors.append("The length of multivariate_support must match the columns of multivariate_counts_map.")
        return errors if errors else True

def _perform_dimensionality_reduction(
        data: pd.DataFrame,
        dim_reduction: str,
        n_components: int,
        verbose: bool = True,
        **reduction_kwargs) -> pd.DataFrame:

    reduction_strategies = {
        'PCA': prince.PCA,
        'MCA': prince.MCA,
        'FAMD': prince.FAMD
    }

    MethodClass = reduction_strategies[dim_reduction]

    if 'scale' in reduction_kwargs:
        if dim_reduction == 'PCA':
            scale_value = reduction_kwargs.pop('scale')
            reduction_kwargs['rescale_with_mean'] = scale_value
            reduction_kwargs['rescale_with_std'] = scale_value
        else:
            reduction_kwargs.pop('scale')

    reduction_method = MethodClass(n_components=n_components, random_state=112, **reduction_kwargs)
    reduced_data = reduction_method.fit_transform(data)
    if verbose:
        print(f'Eigenvalues summary:\n{reduction_method.eigenvalues_summary}')

    return reduced_data

def _scatter_plot(reduced_data: pd.DataFrame, dim_reduction: str, verbose: bool, **kwargs) -> None:
    if verbose:
        warnings.filterwarnings('ignore', category=FutureWarning)
        print(f'Plotting {dim_reduction} 2D Scatter Plot')

    color_col = kwargs.get('color_column', None)
    if color_col is not None and not pd.api.types.is_numeric_dtype(reduced_data[color_col]):
        categories = sorted(reduced_data[color_col].dropna().unique().tolist())
        stops = [i / (len(categories) - 1) if len(categories) > 1 else 0.5 for i in range(len(categories))]
        palette = sample_colorscale('Viridis', stops)
        color_map = {cat: col for cat, col in zip(categories, palette)}

    fig = px.scatter(
        reduced_data,
        x=0,
        y=1,
        title=f'{dim_reduction} Scatter Plot',
        template='plotly_white',
        opacity=0.7,
        color=color_col,
        category_orders={color_col: categories},
        color_discrete_map=color_map
    )

    fig.update_layout(
        title={
            'text': f'{dim_reduction} Scatter Plot',
            'y': 0.95,
            'x': 0.5,
            'xanchor': 'center',
            'yanchor': 'top',
            'font': {'size': 25}
        },
        xaxis_title={
            'text': f'PC1',
            'font': {'size': 18}
        },
        yaxis_title={
            'text': f'PC2',
            'font': {'size': 18}
        }

    )
    fig.show()
    warnings.filterwarnings('default', category=FutureWarning)

def _compute_kde(data_subset, xmin, xmax, kde_resolution):
    """
    Performs a Gaussian Kernel Density Estimation (KDE) on a subset of the original data to estimate
    the probability density function over a specified range, using a grid resolution determined by
    kde_resolution.
    """
    data_subset = np.array(data_subset)
    kde = gaussian_kde(data_subset.T, bw_method='silverman')  # Transpose for data compatibility
    grid = [np.linspace(start, stop, kde_resolution) for start, stop in zip(xmin, xmax)]
    mesh = np.meshgrid(*grid, indexing='ij')
    positions = np.vstack([m.ravel() for m in mesh])
    kde_values = kde(positions).reshape([kde_resolution] * len(xmin))
    return kde_values


def _normalize_kde(kde_values):
    """
    Normalizes the results of the Kernel Density Estimation (KDE) values so that the total area under the
    estimated probability density function equals 1. This ensures that the KDE represents a valid probability
    distribution.
    """
    kde_values = np.maximum(kde_values, 0)  # Set negative values to 0
    return kde_values / np.sum(kde_values)  # Normalize

def _marginalize_multivariate_map(multivariate_map, supports, dimensions) -> list:
    probability_map_list: list = []

    if dimensions == 1:
        probability_map_dim1 = pd.DataFrame(multivariate_map)
        probability_map_list.append(probability_map_dim1)

    elif dimensions == 2:
        probability_map_dim1 = pd.DataFrame([np.sum(dim1, axis=1) for dim1 in multivariate_map],
                                            columns=supports[0])
        probability_map_dim2 = pd.DataFrame([np.sum(dim2, axis=0) for dim2 in multivariate_map],
                                            columns=supports[1])
        probability_map_list.extend([probability_map_dim1, probability_map_dim2])

    elif dimensions == 3:
        probability_map_dim1 = pd.DataFrame([np.sum(dim1, axis=(2, 1)) for dim1 in multivariate_map],
                                            columns=supports[0])
        probability_map_dim2 = pd.DataFrame([np.sum(dim2, axis=(2, 0)) for dim2 in multivariate_map],
                                            columns=supports[1])
        probability_map_dim3 = pd.DataFrame([np.sum(dim3, axis=(0, 1)) for dim3 in multivariate_map],
                                            columns=supports[2])
        probability_map_list.extend([probability_map_dim1, probability_map_dim2, probability_map_dim3])

    return probability_map_list

def _date_to_numeric(data, columns_by_type, verbose):
    # Convert dates to numbers
    if any(columns_by_type['date']):
        date_cols = data.columns[columns_by_type['date']]
        for col in date_cols:
            data[col] = pd.to_datetime(data[col], errors='coerce').astype('int64')
        if verbose:
            print('Converting date columns to numeric for distribution analysis')
    return data




# Copyright 2024 Biomedical Data Science Lab, Universitat Politècnica de València (Spain)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Main function for multi-batch metrics exploration.
"""

from typing import Dict

import plotly.graph_objects as go
import plotly.io as pio
from plotly.colors import get_colorscale

from .arrange_metrics import arrange_performance_metrics

_FONTSIZE = 14


def plot_multibatch_performance(*, metrics: Dict[str, float], metric_name: str) -> None:
    """
    Plots a heatmap visualizing the specified metric for multiple batches of training and test models.

    The function takes a dictionary of metrics and filters them based on the metric identifier.
    It then generates a heatmap where the x-axis represents the test batches,
    the y-axis represents the training batches, and the color scale indicates the
    values of the specified metric.

    The plot is interactive and can be explored (zoomed, hovered, etc.) using Plotly.

    Parameters
    ----------
    metrics : dict
        A dictionary where keys are tuples of (training_batch, test_batch, dataset_type),
        and values are the metric values for the corresponding combination.
        The `dataset_type` should be `'test'` to include the metric in the heatmap.

    metric_name : str
        The name of the metric to visualize.
        The function will filter metrics based on this identifier and only plot those for the 'test' set.
        Regression metric names, when applicable:
            - 'MEAN_ABSOLUTE_ERROR'
            - 'MEAN_SQUARED_ERROR'
            - 'ROOT_MEAN_SQUARED_ERROR'
            - 'R_SQUARED'
        Classification metric names, when applicable:
            - 'AUC_{class_identifier}'
            - 'AUC_MACRO'
            - 'LOGLOSS'
            - 'RECALL_{class_identifier}'
            - 'PRECISION_{class_identifier}'
            - 'F1-SCORE_{class_identifier}'
            - 'ACCURACY'
            - 'RECALL_MACRO'
            - 'RECALL_MICRO'
            - 'RECALL_WEIGHTED'
            - 'PRECISION_MACRO'
            - 'PRECISION_MICRO'
            - 'PRECISION_WEIGHTED'
            - 'F1-SCORE_MACRO'
            - 'F1-SCORE_MICRO'
            - 'F1-SCORE_WEIGHTED'

    Returns
    -------
    None
        This function generates and displays an interactive heatmap using Plotly,
        and does not return any value. The heatmap is displayed directly in the output
        environment (e.g., Jupyter notebook, web browser).

    Raises
    ------
    TypeError
        If the `metrics` parameter is not a dictionary or if `metric_identifier` is not a string.
    """

    # Metrics arrangement
    metrics_test_frame = arrange_performance_metrics(metrics=metrics, metric_name=metric_name)

    # Color scale definition
    colorscale = get_colorscale('RdYlGn')
    if metric_name in ('MEAN_ABSOLUTE_ERROR', 'MEAN_SQUARED_ERROR', 'ROOT_MEAN_SQUARED_ERROR', 'LOGLOSS'):
        colorscale = colorscale[::-1]

    # Plotting using Plotly
    heatmap_data = go.Heatmap(
        z=metrics_test_frame.values,  # Values for the heatmap (reversed rows)
        x=metrics_test_frame.columns,  # Columns as x-axis
        y=metrics_test_frame.index,  # Rows as y-axis
        colorscale=colorscale,  # Color scale
        colorbar=dict(title=metric_name),  # Colorbar label
        hovertemplate="%{y}<br>%{x}: %{z:.3f}",  # Tooltip on hover
        showscale=True  # Display colorbar scale
    )

    # Layout of the plot
    layout = go.Layout(
        title=f'{metric_name.lower().capitalize()} heatmap',
        xaxis=dict(title='Test Batch', tickangle=45, tickfont=dict(size=_FONTSIZE - 2)),
        yaxis=dict(title='Training Batch', tickfont=dict(size=_FONTSIZE - 2)),
        font=dict(size=_FONTSIZE, family="serif"),
        template="plotly_white"  # Optional: use a clean white background template
    )

    # Set the Plotly renderer for Jupyter or standalone use
    # pio.renderers.default = 'notebook'  # For Jupyter Notebooks (use 'notebook' or 'jupyterlab')
    # For standalone (non-Jupyter) use, you can also use:
    #pio.renderers.default = 'browser'

    # Create the figure and plot
    fig = go.Figure(data=[heatmap_data], layout=layout)
    fig.show()



# Copyright 2024 Biomedical Data Science Lab, Universitat Politècnica de València (Spain)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Information Geometric Temporal (IGT) plotting main functions and classes
"""

from datetime import datetime
from typing import Optional

import numpy as np
import plotly.graph_objs as go
import plotly.io as pio
from plotly.colors import sample_colorscale

from dashi._constants import VALID_COLOR_PALETTES, TEMPORAL_PERIOD_YEAR, TEMPORAL_PERIOD_MONTH, TEMPORAL_PERIOD_WEEK, \
    MONTH_LONG_ABBREVIATIONS
from dashi.unsupervised_characterization.variability_metrics.igt_projection import IGTProjection
from dashi.unsupervised_characterization.variability_metrics.igt_trajectory_estimator import _estimate_igt_trajectory
from dashi.utils import _format_date_for_year, _format_date_for_month, _format_date_for_week

pio.renderers.default = 'browser'

def plot_IGT_projection(
        igt_projection: IGTProjection,
        dimensions: int = 2,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        color_palette: str = 'Spectral',
        trajectory: bool = False
) -> go.Figure:
    """
    Plots an interactive Information Geometric Temporal (IGT) plot from an \code{IGTProjection} object.
    An IGT plot visualizes the variability among time batches in a data repository in a 2D or 3D plot.
    Time batches are positioned as points where the distance between them represents the probabilistic
    distance between their distributions (currently Jensen-Shannon distance).
    To track the temporal evolution, temporal batches are labeled to show their date and
    colored according to their season or period, according to the analysis period, as follows.
    If period=="year" the label is "yy" (2 digit year) and the color is according to year.
    If period=="month" the label is "yym" (yy + abbreviated month*) and the color is according
    to the season (yearly).
    If period=="week" the label is "yymmw" (yym + ISO week number in 1-2 digit) and the color is
    according to the season (yearly). An estimated smoothed trajectory of the information evolution
    over time can be shown using the optional "trajectory" parameter.

    Note that since the projection is based on multidimensional scaling, a 2 dimensional
    projection entails a loss of information compared to a 3 dimensional projection. E.g., periodic
    variability components such as seasonal effect can be hindered by an abrupt change or a general trend.

    Parameters
    ----------
    igt_projection : IGTProjection
        The `IGTProjection` object containing the data for the temporal plot.

    dimensions : int, optional
        The number of dimensions to be used for plotting the IGT projection (2D or 3D). Default is 2.

    start_date : Optional[datetime], optional
        The starting date for the temporal plot. If None, it is not constrained. Default is None.

    end_date : Optional[datetime], optional
        The ending date for the temporal plot. If None, it is not constrained. Default is None.

    color_palette : PlotColorPalette, optional
        The color palette to be used for coloring the points. Default is Spectral.

    trajectory : bool, optional
        If True, a smoothed trajectory showing the evolution of the information across time is plotted.
        Default is False.

    Returns
    -------
    Figure
        The Plotly figure object containing the IGT projection plot.
    """
    if dimensions not in [2, 3]:
        raise ValueError(
            'Currently IGT plot can only be made on 2 or 3 dimensions, please set dimensions parameter accordingly')

    if color_palette not in VALID_COLOR_PALETTES:
        raise ValueError(f'color_palette must be one of the defined in {VALID_COLOR_PALETTES}')

    if not start_date:
        start_date = min(igt_projection.data_temporal_map.dates)
    if not end_date:
        end_date = max(igt_projection.data_temporal_map.dates)

    # Date filtering
    date_mask = (igt_projection.data_temporal_map.dates >= np.datetime64(start_date)) & (
            igt_projection.data_temporal_map.dates <= np.datetime64(end_date))
    dates = igt_projection.data_temporal_map.dates[date_mask]
    projection = igt_projection.projection[date_mask]

    # Estimating trajectory if needed
    if trajectory:
        igt_trajectory = _estimate_igt_trajectory(igt_projection)
        trajectory_points = igt_trajectory['points']
        trajectory_dates = igt_trajectory['dates']

    # Generate colors for ten data points
    # Set color based on period
    period = igt_projection.data_temporal_map.period
    colors = []
    period_colors = []

    if period == TEMPORAL_PERIOD_YEAR:
        colors = sample_colorscale(color_palette, [i / len(dates) for i in range(len(dates) + 1)])
    elif period in [TEMPORAL_PERIOD_MONTH, TEMPORAL_PERIOD_WEEK]:
        color_list = sample_colorscale(color_palette, [i / (128 - 1) for i in range(128)])
        color_list.reverse()

        days_of_period = 12 if period == TEMPORAL_PERIOD_MONTH else 53

        color_list.extend(reversed(color_list))
        colors = np.array(color_list)

        period_indexes = np.round(np.linspace(0, 255, days_of_period)).astype(int)
        period_colors = colors.take([period_indexes])[0]

    fig = go.Figure()

    # Plotting
    if dimensions == 2:
        if period == TEMPORAL_PERIOD_YEAR:
            # Add scatter for each point
            for i, (x, y) in enumerate(projection):
                fig.add_trace(
                    go.Scatter(
                        x=[x],
                        y=[y],
                        mode='text',
                        hoverinfo='text',
                        marker=dict(
                            color=colors[i]
                        ),
                        text=_format_date_for_year(dates[i]),
                        textposition="top center",
                        textfont_color=colors[i]
                    )
                )
        elif period == TEMPORAL_PERIOD_MONTH:
            # Add scatter for each point
            for i, (x, y) in enumerate(projection):
                fig.add_trace(
                    go.Scatter(
                        x=[x],
                        y=[y],
                        mode='text',
                        hoverinfo='text',
                        marker=dict(
                            color=period_colors[dates[i].month - 1]
                        ),
                        hovertext=f"{dates[i].strftime('%Y')}-{MONTH_LONG_ABBREVIATIONS[dates[i].month - 1]}",
                        text=_format_date_for_month(dates[i]),
                        textposition="top center",
                        textfont_color=period_colors[dates[i].month - 1]
                    )
                )
        elif period == TEMPORAL_PERIOD_WEEK:
            # Add scatter for each point
            for i, (x, y) in enumerate(projection):
                fig.add_trace(
                    go.Scatter(
                        x=[x],
                        y=[y],
                        mode='text',
                        hoverinfo='text',
                        marker=dict(
                            color=period_colors[dates[i].isoweekday() - 1]
                        ),
                        text=_format_date_for_week(dates[i]),
                        textposition="top center",
                        textfont_color=period_colors[dates[i].isoweekday() - 1]
                    )
                )

        # Add trajectory if necessary
        if trajectory:
            fig.add_trace(
                go.Scatter(
                    x=trajectory_points['D1'],
                    y=trajectory_points['D2'],
                    mode='lines',
                    line=dict(color="#21908C", width=1),
                    hovertext=[f"Approx. date: {date}" for date in trajectory_dates]
                )
            )

        title = {
            'text': 'Information Geometric Temporal (IGT) projection',
            'x': 0.5,
            'y': 0.95,
            'xanchor': 'center',
            'yanchor': 'top',
            'font': {'color': 'black'}
        }

    elif dimensions == 3:
        if period == TEMPORAL_PERIOD_YEAR:
            # Add scatter for each point
            for i, (x, y, z) in enumerate(projection):
                fig.add_trace(
                    go.Scatter3d(
                        x=[x],
                        y=[y],
                        z=[z],
                        mode='text',
                        hoverinfo='text',
                        marker=dict(
                            color=colors[i]
                        ),
                        text=_format_date_for_year(dates[i]),
                        textposition="top center",
                        textfont_color=colors[i]
                    )
                )
        elif period == TEMPORAL_PERIOD_MONTH:
            # Add scatter for each point
            for i, (x, y, z) in enumerate(projection):
                fig.add_trace(
                    go.Scatter3d(
                        x=[x],
                        y=[y],
                        z=[z],
                        mode='text',
                        hoverinfo='text',
                        hovertext=f"{dates[i].strftime('%Y')}-{MONTH_LONG_ABBREVIATIONS[dates[i].month - 1]}",
                        marker=dict(
                            color=period_colors[dates[i].month - 1]
                        ),
                        text=_format_date_for_month(dates[i]),
                        textposition="top center",
                        textfont_color=period_colors[dates[i].month - 1]
                    )
                )
        elif period == TEMPORAL_PERIOD_WEEK:
            # Add scatter for each point
            for i, (x, y, z) in enumerate(projection):
                fig.add_trace(
                    go.Scatter3d(
                        x=[x],
                        y=[y],
                        z=[z],
                        mode='text',
                        hoverinfo='text',
                        marker=dict(
                            color=period_colors[dates[i].isoweekday() - 1]
                        ),
                        text=_format_date_for_week(dates[i]),
                        textposition="top center",
                        textfont_color=period_colors[dates[i].isoweekday() - 1]
                    )
                )
        # Add trajectory if necessary
        if trajectory:
            fig.add_trace(
                go.Scatter3d(
                    x=trajectory_points['D1'],
                    y=trajectory_points['D2'],
                    z=trajectory_points['D3'],
                    mode='lines',
                    line=dict(
                        color="#21908C", width=1.3,  # Color based on row index
                        showscale=False
                    ),
                    hovertext=[f"Approx. date: {date}" for date in trajectory_dates]
                )
            )

        title = {
            'text': 'Information Geometric Temporal (IGT) projection',
            'x': 0.5,
            'y': 0.9,
            'xanchor': 'center',
            'yanchor': 'top',
            'font': {'color': 'black'}
        }

    fig.update_layout(
        plot_bgcolor='white',
        paper_bgcolor='white',
        showlegend=False,
        title=title,
        margin=dict(
            l=40,
            r=40,
            b=50,
            t=90
        ),
        scene=dict(
            xaxis=dict(
                title='D1',
                backgroundcolor="rgba(0, 0, 0, 0)",
                gridcolor="lightgrey",
                showbackground=True,
                zerolinecolor="black",
                titlefont=dict(color='black'),
                tickfont=dict(color='black')
            ),
            yaxis=dict(
                title='D2',
                backgroundcolor="rgba(0, 0, 0, 0)",
                gridcolor="lightgrey",
                showbackground=True,
                zerolinecolor="black",
                titlefont=dict(color='black'),
                tickfont=dict(color='black')
            ),
            zaxis=dict(
                title='D3',
                backgroundcolor="rgba(0, 0, 0, 0)",
                gridcolor="lightgrey",
                showbackground=True,
                zerolinecolor="black",
                titlefont=dict(color='black'),
                tickfont=dict(color='black')
            ),
        ),
    )
    fig.update_xaxes(
        title='D1',
        mirror=True,
        ticks='outside',
        showline=True,
        gridcolor='lightgrey',
        zerolinecolor='black',
        titlefont=dict(color='black'),
        tickfont=dict(color='black')
    )
    fig.update_yaxes(
        title='D2',
        mirror=True,
        ticks='outside',
        showline=True,
        gridcolor='lightgrey',
        zerolinecolor='black',
        titlefont=dict(color='black'),
        tickfont=dict(color='black')
    )
    fig.show()
    return fig



# Copyright 2024 Biomedical Data Science Lab, Universitat Politècnica de València (Spain)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Main module for plotting Multi Source Variability (MSV) metrics.
"""
from dashi.unsupervised_characterization.variability_metrics.estimate_msv_metrics import MSVMetrics
import plotly.graph_objects as go
import plotly.express as px
import plotly.io as pio
from dashi._constants import VALID_COLOR_PALETTES

pio.renderers.default = "browser"

def plot_MSV(msv_metrics: MSVMetrics,
             dimensions: int=1,
             color_palette: str='Spectral'
             ) -> go.Figure:
    """
    Plots a Multi Source Variability (MSV) metrics visualization from a \code{MSVMetrics} object.

    Parameters
    ----------
    msv_metrics : MSVMetrics
        An instance of the `MSVMetrics` class containing the metrics to be plotted.

    dimensions : int, optional
        The number of dimensions for the plot. Must be 1, 2, or 3. Default is 1.

    color_palette : str, optional
        The color palette to use for the plot. Must be one of the valid color palettes defined in `VALID_COLOR_PALETTES`.
        Default is 'Spectral'.

    Returns
    -------
    go.Figure
        A Plotly figure object containing the MSV metrics visualization.
    """

    if dimensions not in [1, 2, 3]:
        raise ValueError('Dimensions must be 1, 2, or 3.')

    if dimensions >= len(msv_metrics.sources):
        raise ValueError(f'Dimensions must go from 1 to the number of sources - 1. Number of sources: '
                         f'{len(msv_metrics.sources)}')

    if color_palette not in VALID_COLOR_PALETTES:
        raise ValueError(f'Invalid color palette. Choose from: {VALID_COLOR_PALETTES}')

    vertices = msv_metrics.vertices
    spos = msv_metrics.SPO
    n_by_source = msv_metrics.nBySource
    id_source = msv_metrics.sources

    sphere_max_size = 100
    scale_factor = sphere_max_size / max(n_by_source)

    sizes = [int(x) * scale_factor for x in n_by_source]

    if dimensions == 1:
        fig = go.Figure(
            data=go.Scatter(
                x=vertices[:, 0],
                y=[0] * len(vertices),  # y-coordinates are zero for 1D
                mode='markers+text',
                marker=dict(
                    size=sizes,
                    sizemode='diameter',
                    color=spos,
                    colorscale=color_palette,
                    opacity=0.8,
                    colorbar=dict(title='SPOs')
                ),
                text=id_source,
                textposition='top center',
                hovertemplate='<b>%{text}</b><br>'
                              'x: %{x:.2f}<br>'
                              'SPO: %{marker.color:.2f}<extra></extra>'
            )
        )
        fig.update_layout(
            title='Multi Source Variability (MSV) Metrics',
            xaxis_title='D1',
            yaxis_title='D2',
            margin=dict(l=0, r=0, b=0, t=30),
            template='plotly_white',
        ),
        fig.show()
        return fig

    elif dimensions == 2:
        fig = go.Figure(
            data=go.Scatter(
                x=vertices[:, 0],
                y=vertices[:, 1],
                mode='markers+text',
                marker=dict(
                    size=sizes,
                    sizemode='diameter',
                    color=spos,
                    colorscale=color_palette,
                    opacity=0.8,
                    colorbar=dict(title='SPOs')
                ),
                text=id_source,
                textposition='top center',
                hovertemplate='<b>%{text}</b><br>'
                              'x: %{x:.2f}<br>y: %{y:.2f}<br>'
                              'SPO: %{marker.color:.2f}<extra></extra>'
            )
        )
        fig.update_layout(
            title='Multi Source Variability (MSV) Metrics',
            xaxis_title='D1',
            yaxis_title='D2',
            margin=dict(l=0, r=0, b=0, t=30),
            template='plotly_white',
        ),
        fig.show()
        return fig

    elif dimensions == 3:
        fig = go.Figure(
            data=go.Scatter3d(
                x=vertices[:, 0],
                y=vertices[:, 1],
                z=vertices[:, 2],
                mode='markers+text',
                marker=dict(
                    size=sizes,
                    sizemode='diameter',
                    color=spos,
                    colorscale=color_palette,
                    colorbar=dict(title='SPOs'),
                    opacity=0.8
                ),
                text=id_source,
                textposition='top center',
                hovertemplate='<b>%{text}</b><br>'
                              'x: %{x:.2f}<br>y: %{y:.2f}<br>z: %{z:.2f}<br>'
                              'SPO: %{marker.color:.2f}<extra></extra>'
            )
        )
        fig.update_layout(
            title="Multi-Source Variability plot",
            plot_bgcolor='white',
            paper_bgcolor='white',
            scene=dict(
                xaxis=dict(
                    title='D1',
                    backgroundcolor="rgba(0, 0, 0, 0)",
                    gridcolor="lightgrey",
                    showbackground=True,
                    zerolinecolor="black",
                    titlefont=dict(color='black'),
                    tickfont=dict(color='black')
                ),
                yaxis=dict(
                    title='D2',
                    backgroundcolor="rgba(0, 0, 0, 0)",
                    gridcolor="lightgrey",
                    showbackground=True,
                    zerolinecolor="black",
                    titlefont=dict(color='black'),
                    tickfont=dict(color='black')
                ),
                zaxis=dict(
                    title='D3',
                    backgroundcolor="rgba(0, 0, 0,0)",
                    gridcolor="lightgrey",
                    showbackground=True,
                    zerolinecolor="black",
                    titlefont=dict(color='black'),
                    tickfont=dict(color='black')
                ),
            ),
            margin=dict(l=0, r=0, b=0, t=30)
        ),
        fig.show()
        return fig
